# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Impact analysis."""

import codecs
from dataclasses import dataclass
import logging
import os
import subprocess
import tempfile
import time
import traceback
from typing import Optional

from google.cloud import ndb
import pygit2
import pygit2.enums

from . import ecosystems
from . import repos
from . import models
from . import vulnerability_pb2

TAG_PREFIX = 'refs/tags/'
BRANCH_PREFIX = 'refs/remotes/'

# Limit for writing small entities.
_DATASTORE_BATCH_SIZE = 5000
# Limit for writing large entities that are close to the 1MiB max entity size.
# There is additionally a request size limit of 10MiB for batched put()
# requests. We conservatively set this to slightly under (8 as opposed to 10)
# to leave some more breathing room.
_DATASTORE_LARGE_BATCH_SIZE = 8
_DATASTORE_BATCH_SLEEP = 10


@dataclass
class AffectedResult:
  """The tags, commits and affected ranges of a vulnerability."""

  tags: set[str]
  commits: set[str]
  affected_ranges: list[str]


@dataclass
class AnalyzeResult:
  """Capturing if an analysis has any changes and what those changes are."""

  has_changes: bool
  commits: set[str]


@dataclass
class TagsInfo:
  """A repository's tags and the one considered to be the latest version."""

  tags: set[str]
  latest_tag: str


class ImpactError(Exception):
  """Impact error."""


class RangeCollector:
  """Affected range collector (preserves insertion order)."""

  def __init__(self):
    self.grouped_ranges = {}

  def add(self, introduced_in, fixed_in, affected_in):
    """Add a new commit range."""
    # last_affected is redundant if fixed is available
    if fixed_in and affected_in:
      affected_in = None

    if introduced_in in self.grouped_ranges:
      if fixed_in is None and affected_in is None:
        # New range doesn't add anything new.
        return

      # Remove in-place as we need to preserve insertion order.
      existing_ranges = self.grouped_ranges[introduced_in]
      existing_ranges.append((introduced_in, fixed_in, affected_in))
      for value in existing_ranges.copy():
        # No fixed or last_affected commits
        if value[1] is None and value[2] is None:
          existing_ranges.remove(value)
          continue

        # Existing last_affected range which is no longer necessary. Now that we
        # have a fixed commit, use that instead.
        if fixed_in and value[1] is None and value[2] is not None:
          existing_ranges.remove(value)
    else:
      self.grouped_ranges[introduced_in] = [(introduced_in, fixed_in,
                                             affected_in)]

  def ranges(self):
    """Return a list representing the collected commit ranges."""
    commit_ranges = []
    for grouped_range in self.grouped_ranges.values():
      for commit_range in grouped_range:
        if commit_range not in commit_ranges:
          commit_ranges.append(commit_range)

    return commit_ranges


class RepoAnalyzer:
  """Repository analyzer.

  This provides functionality for analyzing git repos to determine affected
  tags and commits based on GIT ranges from OSV records.

  Attributes:
    detect_cherrypicks: Whether or not we want to try and detect cherrypicks
      for fix and introduced commits at a best effort basis. This can be slow
      for larger repos. This typically implies `consider_all_branches`.
    consider_all_branches: Whether or not we want to consider all branches when
      analyzing affected commits and tags. For this analysis to avoid false
      positives, it's important that the complete set of `introduced` and
      `fixed` commits are provided (including cherrypicks).
  """

  def __init__(self, detect_cherrypicks=True, consider_all_branches=False):
    self.detect_cherrypicks = detect_cherrypicks
    self.consider_all_branches = consider_all_branches

  def get_affected(self,
                   repo: pygit2.Repository,
                   regress_commits: list[str],
                   fix_commits: list[str],
                   limit_commits: list[str] = None,
                   last_affected_commits: list[str] = None):
    """"Get list of affected tags and commits for a bug given regressed and
    fixed commits."""
    affected_commits, affected_ranges, tags = self._get_affected_range(
        repo,
        regress_commits,
        last_affected_commits,
        fix_commits,
        limit_commits=limit_commits)

    return AffectedResult(tags, affected_commits, affected_ranges)

  def _get_affected_range(self,
                          repo: pygit2.Repository,
                          regress_commits: list[str],
                          last_affected_commits: list[str],
                          fix_commits: list[str],
                          limit_commits: list[str] = None):
    """Get affected range."""
    range_collector = RangeCollector()
    commits = set()
    seen_commits = set()
    tags = set()
    commits_to_tags = _get_commit_to_tag_mappings(repo)
    branch_to_limit = {}
    repo_url = None
    if 'origin' in repo.remotes.names():
      repo_url = repo.remotes['origin'].url

    branches = []

    # If `last_affected` is provided at all, we can't detect cherrypicks, as
    # cherry-pick detection does not make sense when it comes to the
    # `last_affected` commit.
    # Similarly, when there are `limit` commits, we don't do cherry pick
    # detection because it implies limiting to specific branches.
    detect_cherrypicks = (
        self.detect_cherrypicks and not limit_commits and
        not last_affected_commits)

    # detect_cherrypicks implies consider_all_branches because it needs to
    # consider all branches to work.
    consider_all_branches = self.consider_all_branches or detect_cherrypicks

    if consider_all_branches:
      # Check all branches for cherrypicked regress/fix commits (sorted for
      # determinism).
      branches = sorted(repo.branches.remote)
    else:
      if limit_commits:
        for limit_commit in limit_commits:
          current_branches = _branches_with_commit(repo, limit_commit)
          for branch in current_branches:
            branch_to_limit[branch] = limit_commit

          branches.extend(current_branches)
      elif last_affected_commits:
        for last_affected_commit in last_affected_commits:
          branches.extend(_branches_with_commit(repo, last_affected_commit))
      elif fix_commits:
        # TODO(ochang): Remove this check. This behaviour should only be keyed
        # on `limit_commits`.
        # If not detecting cherry picks, take only branches that contain the fix
        # commit. Otherwise we may have false positives.
        for fix_commit in fix_commits:
          branches.extend(_branches_with_commit(repo, fix_commit))
      elif regress_commits:
        # If only a regress commit is available, we need to find all branches
        # that it reaches.
        for regress_commit in regress_commits:
          branches.extend(_branches_with_commit(repo, regress_commit))

    # Optimization: pre-compute branches with specified commits in them if
    # we're not doing cherrypick detection.
    branches_with_commits = {}
    if consider_all_branches and not detect_cherrypicks:
      if fix_commits:
        for fix_commit in fix_commits:
          branches_with_commits[fix_commit] = _branches_with_commit(
              repo, fix_commit)

      if regress_commits:
        for regress_commit in regress_commits:
          branches_with_commits[regress_commit] = _branches_with_commit(
              repo, regress_commit)

    seen_unbounded = set()
    for branch in branches:
      ref = BRANCH_PREFIX + branch

      # Get the earliest equivalent commit in the regression range.
      equivalent_regress_commit = None
      for regress_commit in regress_commits:
        logging.info('Finding equivalent regress commit to %s in %s in %s',
                     regress_commit, ref, repo_url)
        equivalent_regress_commit = self._get_equivalent_commit(
            repo,
            ref,
            regress_commit,
            detect_cherrypicks=detect_cherrypicks,
            branches_with_commits=branches_with_commits)
        if equivalent_regress_commit:
          break

      # If regress_commits is provided, then we should find an equivalent.
      if not equivalent_regress_commit and regress_commits:
        continue

      # Get the latest equivalent commit in the fix range.
      equivalent_fix_commit = None
      for fix_commit in fix_commits:
        logging.info('Finding equivalent fix commit to %s in %s in %s',
                     fix_commit, ref, str(repo_url or 'UNKNOWN_REPO_URL'))
        equivalent_fix_commit = self._get_equivalent_commit(
            repo,
            ref,
            fix_commit,
            detect_cherrypicks=detect_cherrypicks,
            branches_with_commits=branches_with_commits)
        if equivalent_fix_commit:
          break

      # Get the latest equivalent commit in the last_affected range (if
      # present).
      equivalent_last_affected_commit = None
      if last_affected_commits:
        for last_affected_commit in last_affected_commits:
          logging.info(
              'Finding equivalent last_affected commit to %s in %s in %s',
              last_affected_commit, ref, str(repo_url or 'UNKNOWN_REPO_URL'))
          equivalent_last_affected_commit = self._get_equivalent_commit(
              repo,
              ref,
              last_affected_commit,
              # last_affected does not work for cherrypick detection.
              detect_cherrypicks=False)
          if equivalent_last_affected_commit:
            break

      range_collector.add(equivalent_regress_commit, equivalent_fix_commit,
                          equivalent_last_affected_commit)

      if equivalent_fix_commit:
        end_commit = equivalent_fix_commit
        include_end = False
      elif equivalent_last_affected_commit:
        # Note: It's invalid to have both fix and last_affected. In such cases,
        # we prefer the fix due to it coming first in the if statements.
        end_commit = equivalent_last_affected_commit
        include_end = True
      else:
        # Not fixed in this branch. Everything is still vulnerabile.
        end_commit = str(repo.revparse_single(ref).id)
        include_end = True

      if (equivalent_regress_commit, end_commit) in seen_commits:
        continue

      seen_commits.add((equivalent_regress_commit, end_commit))
      cur_commits, cur_tags = get_commit_and_tag_list(
          repo,
          equivalent_regress_commit,
          end_commit,
          commits_to_tags=commits_to_tags,
          include_start=True,
          include_end=include_end,
          limit_commit=branch_to_limit.get(branch),
          seen_unbounded=seen_unbounded)
      commits.update(cur_commits)
      tags.update(cur_tags)

    return commits, range_collector.ranges(), tags

  def _get_equivalent_commit(self,
                             repo,
                             to_search,
                             target_commit,
                             detect_cherrypicks=True,
                             branches_with_commits=None):
    """Find an equivalent commit at to_search, or None. The equivalent commit
    can be equal to target_commit."""
    if not target_commit:
      return None

    # Optimization: If we're not detecting cherrypicks, then we don't need to
    # walk the entire history and we can just look up if a branch contains a
    # commit based on a precomputed dictionary.
    if not detect_cherrypicks and branches_with_commits:
      if (to_search.removeprefix(BRANCH_PREFIX)
          in branches_with_commits.get(target_commit, [])):
        return target_commit

      return None

    try:
      target = repo.revparse_single(target_commit)
    except KeyError:
      # Invalid commit.
      return None

    if detect_cherrypicks:
      try:
        target_patch_id = repo.diff(target.parents[0], target).patchid
      except IndexError:
        # Orphaned target_commit.
        return None
    else:
      target_patch_id = None

    search = repo.revparse_single(to_search)
    try:
      commits = repo.walk(search.id)
    except ValueError:
      # Invalid commit
      return None

    for commit in commits:
      if commit.id == target.id:
        return target_commit

      if not detect_cherrypicks:
        continue

      # Ignore commits without parents and merge commits with multiple parents.
      if not commit.parents or len(commit.parents) > 1:
        continue

      patch_id = repo.cache.get(commit.id)
      if not patch_id:
        diff = repo.diff(commit.parents[0], commit)
        patch_id = diff.patchid
        repo.cache[commit.id] = patch_id

      if patch_id == target_patch_id:
        return str(commit.id)

    # TODO(ochang): Possibly look at commit message, author etc.
    return None


def _get_commit_to_tag_mappings(repo: pygit2.Repository):
  """Get all commit to tag mappings"""
  mappings = {}
  for ref_name in repo.references:
    if not ref_name.startswith(TAG_PREFIX):
      continue

    ref = repo.references[ref_name]
    mappings.setdefault(str(ref.resolve().peel().id),
                        []).append(ref_name[len(TAG_PREFIX):])

  return mappings


def get_commit_and_tag_list(repo,
                            start_commit,
                            end_commit,
                            commits_to_tags=None,
                            include_start=False,
                            include_end=True,
                            limit_commit=None,
                            seen_unbounded: Optional[set[str]] = None):
  """Given a commit range, return the list of commits and tags in the range."""
  if limit_commit:
    if str(repo.merge_base(end_commit, limit_commit)) == limit_commit:
      # Limit commit is an earlier ancestor, so use that as the end of the
      # range instead.
      include_end = False
      end_commit = limit_commit

  repo_url = None
  if 'origin' in repo.remotes.names():
    repo_url = repo.remotes['origin'].url

  logging.info('Getting commits %s..%s from %s', start_commit, end_commit,
               str(repo_url or 'UNKNOWN_REPO_URL'))
  try:
    walker = repo.walk(
        end_commit,
        pygit2.enums.SortMode.TOPOLOGICAL | pygit2.enums.SortMode.REVERSE)
  except KeyError as e:
    raise ImpactError('Invalid commit.') from e

  if start_commit:
    walker.hide(start_commit)

  commits = []
  tags = []

  def process_commit(commit):
    # Optimisation: If we've walked through a commit before and it wasn't bound
    # to a start commit (i.e. affected from the very beginning of time), then
    # record that so we don't have to repeatedly walk through this commit in
    # other branches.
    if not start_commit and seen_unbounded is not None:
      seen_unbounded.add(commit)

    commits.append(commit)
    if not commits_to_tags:
      return

    tags.extend(commits_to_tags.get(commit, []))

  for commit in walker:
    if not include_end and str(commit.id) == end_commit:
      continue

    # Another walker has encountered this commit already, and it was unbounded
    # so we don't need to walk through this again.
    if seen_unbounded and str(commit.id) in seen_unbounded:
      walker.hide(commit.id)
      for parent in commit.parents:
        walker.hide(parent.id)

    process_commit(str(commit.id))

  if include_start and start_commit:
    process_commit(start_commit)

  return commits, tags


def _branches_with_commit(repo, commit):
  """Get all remote branches that include a commit."""
  # pygit2's implementation of this is much slower, so we use `git`.
  try:
    branches = subprocess.check_output(
        ['git', '-C', repo.path, 'branch', '-r', '--contains',
         commit]).decode().splitlines()
  except subprocess.CalledProcessError:
    branches = []

  def process_ref(ref):
    return ref.strip().split()[0]

  # Ignore duplicate <remote>/HEAD branch.
  return [process_ref(branch) for branch in branches if '/HEAD' not in branch]


def _batcher(entries, batch_size):
  """Batcher."""
  for i in range(0, len(entries), batch_size):
    yield entries[i:i + batch_size], i + batch_size >= len(entries)


def _throttled_put(to_put, batch_size=_DATASTORE_BATCH_SIZE):
  """Throttled ndb put."""
  for batch, is_last in _batcher(to_put, batch_size):
    ndb.put_multi(batch)
    if not is_last:
      time.sleep(_DATASTORE_BATCH_SLEEP)


def _throttled_delete(to_delete, batch_size=_DATASTORE_BATCH_SIZE):
  """Throttled ndb delete."""
  for batch, is_last in _batcher(to_delete, batch_size):
    ndb.delete_multi(batch)
    if not is_last:
      time.sleep(_DATASTORE_BATCH_SLEEP)


def update_affected_commits(bug_id, commits, public):
  """Update affected commits."""
  to_put = []
  to_delete = []

  # Write batched commit indexes.
  # Sort the commits for ordering consistency in tests.
  num_pages = 0
  for batch, _ in _batcher(
      sorted(commits), models.AffectedCommits.MAX_COMMITS_PER_ENTITY):
    affected_commits = models.AffectedCommits(
        id=f'{bug_id}-{num_pages}',
        bug_id=bug_id,
        public=public,
        page=num_pages)
    affected_commits.commits = [
        codecs.decode(commit, 'hex') for commit in batch
    ]
    to_put.append(affected_commits)
    num_pages += 1

  # Clear any previously written pages above our current page count.
  for existing in models.AffectedCommits.query(
      models.AffectedCommits.bug_id == bug_id):
    if existing.page >= num_pages:
      to_delete.append(existing.key)

  _throttled_put(to_put, batch_size=_DATASTORE_LARGE_BATCH_SIZE)
  _throttled_delete(to_delete, batch_size=_DATASTORE_LARGE_BATCH_SIZE)


def delete_affected_commits(bug_id):
  """Delete affected commits."""
  affected_commits = models.AffectedCommits.query(
      models.AffectedCommits.bug_id == bug_id)
  _throttled_delete([commit.key for commit in affected_commits],
                    batch_size=_DATASTORE_LARGE_BATCH_SIZE)


def enumerate_versions(package, ecosystem, affected_range):
  """Enumerate versions from SEMVER and ECOSYSTEM input ranges."""
  versions = set()
  sorted_events = []
  limits = []

  # Remove any magic '0' values.
  zero_event = None
  for event in affected_range.events:
    if event.introduced == '0':
      zero_event = event
      continue

    if event.introduced or event.fixed or event.last_affected:
      sorted_events.append(event)
      continue

    if event.limit:
      limits.append(event.limit)

  def sort_key(event):
    """Sort key."""
    if event.introduced:
      return ecosystem.sort_key(event.introduced)
    if event.fixed:
      return ecosystem.sort_key(event.fixed)
    if event.last_affected:
      return ecosystem.sort_key(event.last_affected)

    raise ValueError('Invalid event')

  sorted_events.sort(key=sort_key)
  if zero_event:
    sorted_events.insert(0, zero_event)

  last_introduced = None
  for event in sorted_events:
    if event.introduced and not last_introduced:
      last_introduced = event.introduced

    if last_introduced and event.fixed:
      current_versions = ecosystem.enumerate_versions(
          package, last_introduced, fixed=event.fixed, limits=limits)
      if current_versions:
        versions.update(current_versions)
      last_introduced = None

    if last_introduced and event.last_affected:
      current_versions = ecosystem.enumerate_versions(
          package,
          last_introduced,
          last_affected=event.last_affected,
          limits=limits)
      if current_versions:
        versions.update(current_versions)
      last_introduced = None

  if last_introduced:
    current_versions = ecosystem.enumerate_versions(
        package, last_introduced, limits=limits)
    if current_versions:
      versions.update(current_versions)

  versions = list(versions)
  ecosystem.sort_versions(versions)
  return versions


def _analyze_git_ranges(repo_analyzer: RepoAnalyzer, checkout_path: str,
                        affected_range: vulnerability_pb2.Range,
                        new_versions: set, commits: set, new_introduced: set,
                        new_fixed: set) -> tuple[set, set]:
  """Analyze Git ranges.

  Args:
    repo_analyzer: an instance of RepoAnalyzer to use.
    checkout_path: If defined, used in lieu of cloning the repo.
    affected_range: the GIT range from the vulnerability.
    new_versions: a set that will be in-place modified to contain any new
    versions detected by analysis.
    commits: a set that will be in-place modified to contain any commits???
    new_introduced: a set that will be in-place modified to contain additional
    introduced commits determined by cherry-pick detection.
    new_fixed: a set that will be in-place modified to contain additional fixed
    commits determined determined by cherry-pick detection.

  Returns:
    A tuple of the set of new_versions and commits
  """
  package_repo = None

  with tempfile.TemporaryDirectory() as package_repo_dir:
    # We'd prefer to only download the commit history, but detect_cherrypicks
    # requires having the blobs available to function correctly.
    blobless = not repo_analyzer.detect_cherrypicks
    if checkout_path:
      repo_name = os.path.basename(
          affected_range.repo.rstrip('/')).rstrip('.git')
      package_repo = repos.ensure_updated_checkout(
          affected_range.repo,
          os.path.join(checkout_path, repo_name),
          blobless=blobless)
    else:
      package_repo = repos.clone_with_retries(
          affected_range.repo, package_repo_dir, blobless=blobless)

    all_introduced = []
    all_fixed = []
    all_last_affected = []
    all_limit = []
    for event in affected_range.events:
      if event.introduced and event.introduced != '0':
        all_introduced.append(event.introduced)
        continue

      if event.last_affected:
        all_last_affected.append(event.last_affected)
        continue

      if event.fixed:
        all_fixed.append(event.fixed)
        continue

      if event.limit:
        all_limit.append(event.limit)
        continue

    try:
      result = repo_analyzer.get_affected(package_repo, all_introduced,
                                          all_fixed, all_limit,
                                          all_last_affected)
    except ImpactError:
      logging.warning('Got error while analyzing git range in %s: %s',
                      affected_range.repo, traceback.format_exc())
      return new_versions, commits

    for introduced, fixed, _ in result.affected_ranges:
      if introduced and introduced not in all_introduced:
        new_introduced.add(introduced)

      if fixed and fixed not in all_fixed:
        new_fixed.add(fixed)

    new_versions.update(result.tags)
    commits.update(result.commits)

  return new_versions, commits


def analyze(vulnerability: vulnerability_pb2.Vulnerability,
            analyze_git: bool = True,
            checkout_path: str = None,
            detect_cherrypicks: bool = True,
            versions_from_repo: bool = True,
            consider_all_branches: bool = False) -> AnalyzeResult:
  """Analyze and possibly update a vulnerability based on its input ranges.

  The behaviour varies by the vulnerability's affected field.

  If there's package information for a supported ecosystem, versions may be
  enumerated.
  If there's GIT ranges and analyze_git and versions_from_repo are True,
  versions are enumerated from the associated Git repo.

  Args:
    vulnerability: A vulnerability_pb2.Vulnerability message.
    analyze_git: If True and there is a GIT range, the related repo is analyzed
    further.
    checkout_path: If defined, used in lieu of cloning the repo.
    detect_cherrypicks: If True, cherrypick detection is performed during repo
    analysis.
    versions_from_repo: add the versions derived from the Git repo to the
    affected.versions field.

  Returns:
    An AnalyzeResult dataclass, indicating if anything changed the relevant
    commits.
  """
  commits = set()
  has_changes = False
  for affected in vulnerability.affected:
    versions = []
    for affected_range in affected.ranges:
      if (affected_range.type == vulnerability_pb2.Range.ECOSYSTEM and
          ecosystems.is_semver(affected.package.ecosystem)):
        # Replace erroneous range type.
        affected_range.type = vulnerability_pb2.Range.SEMVER

      if affected_range.type in (vulnerability_pb2.Range.ECOSYSTEM,
                                 vulnerability_pb2.Range.SEMVER):
        # Enumerate ECOSYSTEM and SEMVER ranges.
        ecosystem_helpers = ecosystems.get(affected.package.ecosystem)
        if ecosystem_helpers is None:
          logging.warning('No ecosystem helpers implemented for %s: %s',
                          affected.package.ecosystem, vulnerability.id)
        elif isinstance(ecosystem_helpers, ecosystems.EnumerableEcosystem):
          try:
            versions.extend(
                enumerate_versions(affected.package.name, ecosystem_helpers,
                                   affected_range))
          except ecosystems.EnumerateError:
            # Allow non-retryable enumeration errors to occur (e.g. if the
            # package no longer exists).
            pass

      new_git_versions = set()
      new_introduced = set()
      new_fixed = set()

      # Analyze git ranges.
      if (analyze_git and
          affected_range.type == vulnerability_pb2.Range.Type.GIT):
        repo_analyzer = RepoAnalyzer(
            detect_cherrypicks=detect_cherrypicks,
            consider_all_branches=consider_all_branches)
        try:
          _analyze_git_ranges(repo_analyzer, checkout_path, affected_range,
                              new_git_versions, commits, new_introduced,
                              new_fixed)
        except Exception as e:
          e.add_note(f'Happened analyzing {vulnerability.id}')
          raise

      # Add additional versions derived from commits and tags.
      if versions_from_repo:
        versions.extend(new_git_versions)

      # Apply changes.
      for introduced in new_introduced:
        if (not any(
            event.introduced == introduced for event in affected_range.events)):
          has_changes = True
          affected_range.events.add(introduced=introduced)

      for fixed in new_fixed:
        if not any(event.fixed == fixed for event in affected_range.events):
          has_changes = True
          affected_range.events.add(fixed=fixed)

    for version in sorted(versions):
      if version not in affected.versions:
        has_changes = True
        affected.versions.append(version)

  if not has_changes:
    return AnalyzeResult(has_changes=False, commits=commits)

  vulnerability.modified.FromDatetime(models.utcnow())
  return AnalyzeResult(has_changes=True, commits=commits)



# Copyright 2021 OSV Schema Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the GHSA to OSV converter."""

import json
import os
import unittest

import convert_ghsa

TEST_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'testdata')

class ConverterTest(unittest.TestCase):
    """Converter unit tests."""

    def setUp(self):
        self.maxDiff = None  # pylint: disable=invalid-name

    def check_conversion(self, name):
        """Check OSV conversation against the expected result."""
        with open(os.path.join(TEST_DIR, f'{name}.json')) as handle:
            input_data = json.load(handle)

        output = convert_ghsa.convert(input_data)

        expected_path = os.path.join(TEST_DIR, f'{name}.osv.json')
        if os.getenv('TESTS_GENERATE'):
            with open(expected_path, 'w') as handle:
                handle.write(json.dumps(output, indent=2))

        with open(expected_path) as handle:
            expected = json.load(handle)

        self.assertDictEqual(expected, output)

    def test_full_ranges(self):
        """Test full (>= X, < Y) ranges."""
        self.check_conversion('full_ranges')

    def test_greater_than_equals_no_patch(self):
        """Test >=X without a patch."""
        self.check_conversion('greater_than_equals_no_patch')

    def test_multiple_ranges_in_package(self):
        """Test multiple ranges in the same package."""
        self.check_conversion('multiple_ranges_in_package')

    def test_less_than_equals_no_patch(self):
        """Test less than equals with no patch."""
        self.check_conversion('less_than_equals_no_patch')

    def test_less_than_equals_with_patch(self):
        """Test less than equals with patch."""
        self.check_conversion('less_than_equals_with_patch')

    def test_equals_no_patch(self):
        """Test =X versions with no patch."""
        self.check_conversion('equals_no_patch')

    def test_equals_with_patch(self):
        """Test =X versions with a patch."""
        self.check_conversion('equals_with_patch')

    def test_withdrawn(self):
        """Test a withdrawn entry."""
        self.check_conversion('withdrawn')

    def test_pypi_normalize(self):
        """Test normalization PyPI names."""
        self.check_conversion('pypi_normalize')

    def test_maven_greater_than(self):
        """Test Maven > ranges."""
        self.check_conversion('maven_greater_than')

    def test_maven_greater_than(self):
        """Test npm > ranges."""
        self.check_conversion('npm_greater_than')



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Repo functions."""
import datetime
import logging
import os
import shutil
import subprocess
import time

import pygit2
import pygit2.enums

CLONE_TRIES = 3
RETRY_SLEEP_SECONDS = 5

# More performant mirrors for large/popular repos.
# TODO: Don't hardcode this.
_GIT_MIRRORS = {
    'https://git.kernel.org/pub/scm/linux/kernel/git/stable/linux.git':
        'https://kernel.googlesource.com/pub/scm/'
        'linux/kernel/git/stable/linux.git'
}

FETCH_CACHE: dict[tuple, datetime.datetime] = {}
FETCH_CACHE_SECONDS = 5 * 60  # 5 minutes


class GitRemoteCallback(pygit2.RemoteCallbacks):
  """Authentication callbacks."""

  def __init__(self, username, ssh_key_public_path, ssh_key_private_path):
    super().__init__()
    self.username = username
    self.ssh_key_public_path = ssh_key_public_path
    self.ssh_key_private_path = ssh_key_private_path

  def credentials(self, url, username_from_url, allowed_types):
    """Get credentials."""
    del url
    del username_from_url
    if allowed_types & pygit2.enums.CredentialType.USERNAME:
      return pygit2.Username(self.username)

    if allowed_types & pygit2.CredentialType.SSH_KEY:
      return pygit2.Keypair(self.username, self.ssh_key_public_path,
                            self.ssh_key_private_path, '')

    return None


def _git_mirror(git_url):
  """Get git mirror. If no mirror exists, return the git URL as is."""
  mirror = _GIT_MIRRORS.get(git_url.rstrip('/'))
  if mirror:
    logging.info('Using mirror %s for git URL %s.', mirror, git_url)
    return mirror

  return git_url


def _checkout_branch(repo, branch):
  """Check out a branch."""
  remote_branch = repo.lookup_branch('origin/' + branch,
                                     pygit2.enums.BranchType.REMOTE)
  local_branch = repo.lookup_branch(branch, pygit2.enums.BranchType.LOCAL)
  if not local_branch:
    if remote_branch is None:
      raise NoBranchError
    local_branch = repo.branches.create(branch, commit=remote_branch.peel())

  local_branch.upstream = remote_branch
  local_branch.set_target(remote_branch.target)
  repo.checkout(local_branch)
  repo.reset(remote_branch.target, pygit2.enums.ResetMode.HARD)


def _set_git_callback_env(git_callbacks):
  """Set the environment variable to set git callbacks for cli git"""
  env = {
      # Prevent prompting for username if we don't have an ssh key
      'GIT_TERMINAL_PROMPT': '0'
  }
  if git_callbacks:
    env['GIT_SSH_COMMAND'] = (
        f'ssh -i "{git_callbacks.ssh_key_private_path}" '
        f'-o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null '
        f'-o User={git_callbacks.username} -o IdentitiesOnly=yes')
  return env


class GitCloneError(Exception):
  """Git repository clone exception."""


class NoBranchError(Exception):
  """Branch does not exist"""


def clone(git_url, checkout_dir, git_callbacks=None, blobless=False):
  """Perform a clone."""
  try:
    # Use 'git' CLI here as it's much faster than libgit2's clone.
    env = _set_git_callback_env(git_callbacks)
    cmd = ['git', 'clone']
    if blobless:
      cmd.append('--filter=blob:none')
    cmd.extend([_git_mirror(git_url), checkout_dir])
    subprocess.run(cmd, env=env, capture_output=True, check=True)
    return pygit2.Repository(checkout_dir)
  except subprocess.CalledProcessError as e:
    raise GitCloneError(f'Failed to clone repo:\n{e.stderr.decode()}') from e
  except pygit2.GitError as e:
    raise GitCloneError('Failed to open cloned repo') from e


def clone_with_retries(git_url,
                       checkout_dir,
                       git_callbacks=None,
                       branch=None,
                       blobless=False):
  """Clone with retries."""
  logging.info('Cloning %s to %s', git_url, checkout_dir)
  for attempt in range(CLONE_TRIES):
    try:
      repo = clone(git_url, checkout_dir, git_callbacks, blobless=blobless)
      repo.cache = {}
      if branch:
        _checkout_branch(repo, branch)
      return repo
    except GitCloneError as e:
      shutil.rmtree(checkout_dir, ignore_errors=True)
      if attempt == CLONE_TRIES - 1:
        raise GitCloneError('Clone failed after %d attempts' %
                            CLONE_TRIES) from e
      time.sleep(RETRY_SLEEP_SECONDS)
    except NoBranchError as e:
      raise NoBranchError('Branch "%s" not found in repo "%s"' %
                          (branch, git_url)) from e

  return None


def _use_existing_checkout(git_url,
                           checkout_dir,
                           git_callbacks=None,
                           branch=None):
  """Update and use existing checkout."""
  repo = pygit2.Repository(checkout_dir)
  repo.cache = {}
  if repo.remotes['origin'].url != _git_mirror(git_url):
    # The URL in the code is the source of truth,
    # so if the remote URL does not match
    # update to the correct URL.
    logging.warning('origin URL updated:\nOld: %s\nNew: %s',
                    repo.remotes['origin'].url, _git_mirror(git_url))
    repo.remotes['origin'].url = _git_mirror(git_url)

  if branch:
    try:
      _checkout_branch(repo, branch)
    except NoBranchError as e:
      raise NoBranchError('Branch "%s" not found in repo "%s"' %
                          (branch, git_url)) from e

  reset_repo(repo, git_callbacks)
  logging.info('Using existing checkout at %s', checkout_dir)
  return repo


def ensure_updated_checkout(git_url,
                            checkout_dir,
                            git_callbacks=None,
                            branch=None,
                            blobless=False):
  """Ensure updated checkout."""
  if os.path.exists(checkout_dir):
    # Already exists, reset and checkout latest revision.
    try:
      return _use_existing_checkout(
          git_url, checkout_dir, git_callbacks=git_callbacks, branch=branch)
    except Exception as e:
      # Failed to re-use existing checkout. Delete it and start over.
      err_str = str(e)
      if isinstance(e, subprocess.CalledProcessError):
        # add the git output to the log
        err_str = f'{err_str}\n{e.stderr.decode()}'
      logging.exception('Failed to load existing checkout: %s', err_str)
      shutil.rmtree(checkout_dir)

  repo = clone_with_retries(
      git_url,
      checkout_dir,
      git_callbacks=git_callbacks,
      branch=branch,
      blobless=blobless)
  logging.info('Repo now at: %s', repo.head.peel().message)
  return repo


def reset_repo(repo: pygit2.Repository, git_callbacks, force: bool = False):
  """
  Fetch the latest changes from remote, and set upstream branch correctly.
  This will try to be smart and not refetch repos that recently have been
  fetched.

  Use force to override this
  """
  remote_url = repo.remotes['origin'].url
  key = (remote_url, repo.path)
  now = datetime.datetime.now(datetime.timezone.utc)

  if not force and key in FETCH_CACHE and (
      now - FETCH_CACHE[key]).total_seconds() < FETCH_CACHE_SECONDS:
    logging.info('Skipping fetch for %s, fetched recently.', remote_url)
    repo.reset(repo.head.target, pygit2.enums.ResetMode.HARD)
    return

  logging.info('Fetching for %s', remote_url)
  env = _set_git_callback_env(git_callbacks)

  # Use git cli instead of pygit2 for performance
  subprocess.run(['git', 'fetch', 'origin'],
                 cwd=repo.workdir,
                 env=env,
                 capture_output=True,
                 check=True)
  FETCH_CACHE[key] = now

  # Pygit2 equivalent of above call
  # repo.remotes['origin'].fetch(callbacks=git_callbacks)
  remote_branch = repo.lookup_branch(
      repo.head.name.replace('refs/heads/', 'origin/'),
      pygit2.enums.BranchType.REMOTE)

  # Reset to remote branch.
  repo.head.set_target(remote_branch.target)
  repo.reset(remote_branch.target, pygit2.enums.ResetMode.HARD)



#!/usr/bin/env python3
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Importer sources."""
import json
import hashlib
import logging
import os

import jsonschema
import pygit2
import time
import yaml

from google.protobuf import json_format

# pylint: disable=relative-beyond-top-level
from . import cache
from . import repos
from . import vulnerability_pb2

AUTHOR_EMAIL = 'infra@osv.dev'
PUSH_RETRIES = 2
PUSH_RETRY_SLEEP_SECONDS = 10

YAML_EXTENSIONS = ('.yaml', '.yml')
JSON_EXTENSIONS = ('.json',)

DEFAULT_TIMESTAMP = 946684800  # Year 2000

shared_cache = cache.InMemoryCache()


class KeyPathError(Exception):
  """
  The provided key path was not found in the object.

  For example, this can happen with GSD entries where for most vulnerabilities,
  an OSV entry is not published, only the GSD part.
  """


def parse_source_id(source_id):
  """Get the source name and id from source_id."""
  return source_id.split(':', 1)


def repo_path(repo):
  """Return local disk path to repo."""
  # Remove '.git' component.
  return os.path.dirname(repo.path.rstrip(os.sep))


class NoDatesSafeLoader(yaml.SafeLoader):
  """
  Safe YAML loader that removes datetime autoparsing

  PyYAML automatically parses date strings into Python datetime.datetime, which
  will cause multiple issues in other parts of the osv library, including
  automatically failing the json schema verifier.
  """

  @classmethod
  def remove_implicit_resolver(cls, tag_to_remove):
    """
    Remove implicit resolvers for a particular tag

    Takes care not to modify resolvers in super classes.
    """
    if 'yaml_implicit_resolvers' not in cls.__dict__:
      cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()

    for first_letter, mappings in cls.yaml_implicit_resolvers.items():
      cls.yaml_implicit_resolvers[first_letter] = [
          (tag, regexp) for tag, regexp in mappings if tag != tag_to_remove
      ]


NoDatesSafeLoader.remove_implicit_resolver('tag:yaml.org,2002:timestamp')


def _parse_vulnerability_dict(path):
  """Parse a vulnerability file into a dict."""
  with open(path) as f:
    ext = os.path.splitext(path)[1]
    if ext in YAML_EXTENSIONS:
      return yaml.load(f, Loader=NoDatesSafeLoader)

    if ext in JSON_EXTENSIONS:
      return json.load(f)

    raise RuntimeError('Unknown format ' + ext)

  return None


@cache.cached(shared_cache)
def load_schema():
  path = os.path.join(
      os.path.dirname(os.path.abspath(__file__)), 'osv-schema', 'validation',
      'schema.json')
  with open(path, 'r') as schema:
    text = schema.read()
    return json.loads(text)


def parse_vulnerability(path, key_path=None, strict=False):
  """Parse vulnerability YAML/JSON."""
  data = _parse_vulnerability_dict(path)
  return parse_vulnerability_from_dict(data, key_path, strict)


def _parse_vulnerabilities(data, key_path, strict=False):
  """Parse multiple vulnerabilities."""
  if isinstance(data, list):
    return [parse_vulnerability_from_dict(v, key_path, strict) for v in data]

  return [parse_vulnerability_from_dict(data, key_path, strict)]


def parse_vulnerabilities(path, key_path=None, strict=False):
  """Parse vulnerabilities (potentially multiple in a list)."""
  return _parse_vulnerabilities(
      _parse_vulnerability_dict(path), key_path, strict)


def parse_vulnerabilities_from_data(data_text,
                                    extension,
                                    key_path=None,
                                    strict=False):
  """Parse vulnerabilities from data."""
  if extension in YAML_EXTENSIONS:
    data = yaml.load(data_text, Loader=NoDatesSafeLoader)
  elif extension in JSON_EXTENSIONS:
    data = json.loads(data_text)
  else:
    raise RuntimeError('Unknown format ' + extension)

  return _parse_vulnerabilities(data, key_path, strict)


def _get_nested_vulnerability(data, key_path=None):
  """Get nested vulnerability."""
  if key_path:
    try:
      for component in key_path.split('.'):
        data = data[component]
    except KeyError as e:
      raise KeyPathError() from e

  return data


def parse_vulnerability_from_dict(data, key_path=None, strict=False):
  """Parse vulnerability from dict."""
  data = _get_nested_vulnerability(data, key_path)
  try:
    jsonschema.validate(data, load_schema())
  except jsonschema.exceptions.ValidationError as e:
    logging.warning('Failed to validate loaded OSV entry: %s', e.message)
    logging.warning('Invalid data: %s', data)
    if strict:  # Reraise the error if strict
      raise

  vulnerability = vulnerability_pb2.Vulnerability()
  json_format.ParseDict(data, vulnerability, ignore_unknown_fields=True)
  if not vulnerability.id:
    raise ValueError('Missing id field. Invalid vulnerability.')

  return vulnerability


def _yaml_str_representer(dumper, data):
  """YAML str representer override."""
  if '\n' in data:
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
  return dumper.represent_scalar('tag:yaml.org,2002:str', data)


class YamlDumper(yaml.SafeDumper):
  """Overridden dumper to to use | for multiline strings."""


YamlDumper.add_representer(str, _yaml_str_representer)


def vulnerability_to_dict(vulnerability):
  """Convert Vulnerability to a dict."""
  result = json_format.MessageToDict(
      vulnerability, preserving_proto_field_name=True)

  if 'affected' not in result:
    return result

  for affected in result['affected']:
    if any(r.get('type') == 'SEMVER' for r in affected.get('ranges', [])):
      return result

  # If no SemVer ranges, output an empty "versions": [] to conform to the spec.
  for affected in result['affected']:
    if 'versions' not in affected:
      affected['versions'] = []

  return result


def _write_vulnerability_dict(data, output_path,
                              modified_date_timestamp: float):
  """Write a vulnerability dict to disk."""
  with open(output_path, 'w') as f:
    ext = os.path.splitext(output_path)[1]
    if ext in YAML_EXTENSIONS:
      yaml.dump(data, f, sort_keys=True, Dumper=YamlDumper)
    elif ext in JSON_EXTENSIONS:
      json.dump(data, f, indent=2, sort_keys=True)
    else:
      raise RuntimeError('Unknown format ' + ext)

  os.utime(output_path, (modified_date_timestamp, modified_date_timestamp))


def write_vulnerability(vulnerability: vulnerability_pb2.Vulnerability,
                        output_path,
                        key_path=None):
  """Update a vulnerability file on disk."""
  if os.path.exists(output_path):
    data = _parse_vulnerability_dict(output_path)
  else:
    # Set up the expected nesting based on key_path.
    data = {}
    if key_path:
      cur = data
      for component in key_path.split('.'):
        cur[component] = {}
        cur = cur[component]

  vuln_data = _get_nested_vulnerability(data, key_path)
  vuln_data.clear()
  vuln_data.update(vulnerability_to_dict(vulnerability))
  dt_timestamp = vulnerability.modified.seconds
  if dt_timestamp < DEFAULT_TIMESTAMP:
    logging.warning('Record has no modified time: %s', vulnerability.id)
    dt_timestamp = DEFAULT_TIMESTAMP

  _write_vulnerability_dict(data, output_path, dt_timestamp)


def vulnerability_has_range(vulnerability, introduced, fixed):
  """Check if a vulnerability has a range."""
  for affected_range in vulnerability.affects.ranges:
    if affected_range.type != vulnerability_pb2.AffectedRange.Type.GIT:
      continue

    if (affected_range.introduced == introduced and
        affected_range.fixed == fixed):
      return True

  return False


def push_source_changes(repo,
                        commit_message,
                        git_callbacks,
                        expected_hashes=None):
  """Push source changes."""
  repo.index.write()
  tree = repo.index.write_tree()
  author = git_author()
  repo.create_commit(repo.head.name, author, author, commit_message, tree,
                     [repo.head.peel().id])

  for retry_num in range(1 + PUSH_RETRIES):
    try:
      repo.remotes['origin'].push([repo.head.name], callbacks=git_callbacks)
      return True
    except pygit2.GitError as e:
      logging.warning('Failed to push: %s', e)
      if retry_num == PUSH_RETRIES:
        repos.reset_repo(repo, git_callbacks, True)
        return False

      time.sleep(PUSH_RETRY_SLEEP_SECONDS)

      # Try rebasing.
      commit = repo.head.peel()
      repos.reset_repo(repo, git_callbacks, True)

      for path, expected_hash in expected_hashes.items():
        current_hash = sha256(path)
        if current_hash != expected_hash:
          logging.warning(
              'Upstream hash for %s changed (expected=%s vs current=%s)', path,
              expected_hash, current_hash)

          continue

      # Then cherrypick our original commit.
      repo.cherrypick(commit.id)
      if repo.index.conflicts is not None:
        # Conflict. Don't try to resolve.
        repo.state_cleanup()
        repos.reset_repo(repo, git_callbacks, True)
        return False

      # Success, commit and try pushing again.
      tree = repo.index.write_tree()
      repo.create_commit(repo.head.name, commit.author, commit.author,
                         commit.message, tree, [repo.head.peel().id])
      repo.state_cleanup()

  return True


def git_author():
  """Get the git author for commits."""
  return pygit2.Signature('OSV', AUTHOR_EMAIL)


def sha256(file_path):
  """Computes sha256 sum."""
  hasher = hashlib.sha256()
  with open(file_path, 'rb') as f:
    hasher.update(f.read())
  return hasher.hexdigest()


def sha256_bytes(data):
  """Computes sha256sum."""
  hasher = hashlib.sha256()
  hasher.update(data)
  return hasher.hexdigest()


def source_path(source_repo, bug):
  """Get the source path for an osv.Bug."""
  source_name, source_id = parse_source_id(bug.source_id)
  if source_name == 'oss-fuzz' and len(bug.project) > 0:
    path = os.path.join(bug.project[0], bug.id() + source_repo.extension)
    if source_repo.directory_path:
      path = os.path.join(source_repo.directory_path, path)

    return path

  return source_id



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Alpine ecosystem helper."""

import glob
import logging
import os.path
import subprocess
import typing

from ..third_party.univers.alpine import AlpineLinuxVersion

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError
from .ecosystems_base import OrderedEcosystem
from .. import repos
from ..cache import cached


class APK(OrderedEcosystem):
  """Alpine Package Keeper ecosystem helper."""

  def sort_key(self, version):
    if not AlpineLinuxVersion.is_valid(version):
      # If version is not valid, it is most likely an invalid input
      # version then sort it to the last/largest element
      return AlpineLinuxVersion('999999')
    return AlpineLinuxVersion(version)


class Alpine(APK, EnumerableEcosystem):
  """
  Alpine packages ecosystem
  
  Args:
      alpine_release_ver: is used to configure which branch 
      in alpines git repo is used for version enumeration
  """

  # Use github mirror which supports more bandwidth.
  _APORTS_GIT_URL = 'https://github.com/alpinelinux/aports.git'
  _BRANCH_SUFFIX = '-stable'
  _GIT_REPO_PATH = 'version_enum/aports/'
  # Sometimes (2 or 3 packages) APKBUILD files are a bash script and version
  # is actually stored in variables. _kver is the common variable name.
  _PKGVER_ALIASES = ('+pkgver=', '+_kver=')
  _PKGREL_ALIASES = ('+pkgrel=', '+_krel=')

  @property
  def alpine_release_ver(self) -> str:
    return self.suffix if self.suffix is not None else ''

  def get_branch_name(self) -> str:
    return self.alpine_release_ver.lstrip('v') + self._BRANCH_SUFFIX

  @staticmethod
  def _process_git_log(output: str) -> list:
    """
    Takes git log diff output,
    finds all changes to pkgver and pkgrel and outputs that in an unsorted list
    """
    all_versions = set()
    # Filter out the relevant lines
    lines = [
        x for x in output.splitlines()
        if len(x) == 0 or x.startswith(Alpine._PKGVER_ALIASES) or
        x.startswith(Alpine._PKGREL_ALIASES)
    ]
    # Reverse so that it's in chronological order.
    # The following loop also expects this order.
    lines.reverse()

    current_ver = None
    current_rel = None

    def clean_versions(ver: str) -> str:
      ver = ver.split(' #')[0]  # Remove comment lines
      ver = ver.strip(' "\'')  # Remove (occasional) quotes
      ver = ver.removeprefix('r')
      return ver

    for line in lines:
      if not line:  # This line ends a commit block
        # Don't add any versions until the entire commit block is finished
        if not current_ver or not current_rel:
          continue

        current_ver = clean_versions(current_ver)
        current_rel = clean_versions(current_rel)
        # Ignore occasional version that is still not valid.
        if AlpineLinuxVersion.is_valid(current_ver) and current_rel.isdigit():
          all_versions.add(current_ver + '-r' + current_rel)
        else:
          logging.warning('Alpine version "%s" - "%s" is not valid',
                          current_ver, current_rel)
        continue

      # Within a commit block, parse all lines and save it till the end
      # of the block
      x_split = line.split('=', 1)
      if len(x_split) != 2:
        # Skip the occasional invalid versions
        continue

      num = x_split[1]
      if line.startswith(Alpine._PKGVER_ALIASES):
        current_ver = num
        continue

      if line.startswith(Alpine._PKGREL_ALIASES):
        current_rel = num
        continue

    # Return unsorted list of versions (as it's converted from a set)
    return list(all_versions)

  def enumerate_versions(self,
                         package: str,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """We use Alpines aports git repository history to do version enum"""

    if config.work_dir is None:
      logging.error(
          "Tried to enumerate alpine version without workdir being set")
      return []

    get_versions = self._get_versions
    if config.shared_cache:
      get_versions = cached(config.shared_cache)(get_versions)

    versions = get_versions(self.get_branch_name(), package)
    self.sort_versions(versions)

    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)

  @staticmethod
  def _get_versions(branch: str, package: str) -> typing.List[str]:
    """Get all versions for a package from aports repo"""

    # Checkout the repo at the alpine version branch
    checkout_dir = os.path.join(config.work_dir, Alpine._GIT_REPO_PATH)
    repos.ensure_updated_checkout(
        Alpine._APORTS_GIT_URL, checkout_dir, branch=branch)
    directories = glob.glob(
        os.path.join(checkout_dir, '*', package.lower(), 'APKBUILD'),
        recursive=True)

    if len(directories) != 1:
      raise EnumerateError('Cannot find package in aports')

    relative_path = os.path.relpath(directories[0], checkout_dir)

    # Run git log -L to get a list of all changes to the package APKBUILD file
    # Specifically changes to "pkgver=" and "pkgrel="
    regex_test_versions = '\\|'.join(
        ['\\(' + x.removeprefix('+') + '\\)' for x in Alpine._PKGVER_ALIASES])

    regex_test_revisions = '\\|'.join(
        ['\\(' + x.removeprefix('+') + '\\)' for x in Alpine._PKGREL_ALIASES])

    # yapf: disable
    # https://git-scm.com/docs/git-log/2.33.1#Documentation/git-log.txt--Lltstartgtltendgtltfilegt
    stdout_data = subprocess.check_output([
        'git', 'log', '--oneline',
        '-L', '^/' + regex_test_versions  + '/,+1:' + relative_path,
        '-L', '^/' + regex_test_revisions + '/,+1:' + relative_path
    ], cwd=checkout_dir).decode('utf-8')
    # yapf: enable

    versions = Alpine._process_git_log(stdout_data)
    return versions



""" Convert a CSAF document to OSV format
    i.e. https://access.redhat.com/security/data/csaf/v2/advisories/2024/rhsa-2024_4546.json
"""

from setuptools import setup, find_packages

REQUIRES = [
    "jsonschema",
    "requests"
]

setup(
    name="redhat_osv",
    version="1.0.0",
    description="Convert Red Hat CSAF documents to OSV format",
    author_email="jshepher@redhat.com",
    url="",
    keywords=["OSV", "CSAF"],
    install_requires=REQUIRES,
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    packages=["redhat_osv"],
    entry_points={"console_scripts": ["convert_redhat=convert_redhat:main"]},
    long_description="The purpose of this tool is to convert from Red Hat CSAF documents to OSV",
)


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SemVer index tests."""

import unittest

import semver

from . import semver_index


class SemverIndexTests(unittest.TestCase):
  """SemVer index tests."""

  def test_coerce(self):
    """Test coerce."""
    self.assertEqual('1.0.0', semver_index.coerce('1'))
    self.assertEqual('1.0.0', semver_index.coerce('v1'))
    self.assertEqual('1.0.0', semver_index.coerce('v1.0'))
    self.assertEqual('1.0.0', semver_index.coerce('1.0.0'))

    self.assertEqual('1.5.0', semver_index.coerce('1.5'))
    self.assertEqual('1.5.0', semver_index.coerce('v1.5'))
    self.assertEqual('1.5.0', semver_index.coerce('v1.5.0'))
    self.assertEqual('1.5.1', semver_index.coerce('1.5.1'))

    self.assertEqual('', semver_index.coerce(''))
    self.assertEqual('rubbish', semver_index.coerce('rubbish'))
    self.assertEqual('a1.0.0', semver_index.coerce('a1.0.0'))
    self.assertEqual('1.0.0.0', semver_index.coerce('1.0.0.0'))
    self.assertEqual('1.0.0-foo', semver_index.coerce('1.0.0-foo'))
    self.assertEqual('1.0.0-foo', semver_index.coerce('1-foo'))

  def test_normalize(self):
    """Test version normalization."""
    versions = [
        '1.0.0-beta.2+BLAH', '1.0.0-beta.2', '1.0.0-beta.11', '1.0.0-rc.1',
        '1.0.0', '1.0.0-alpha', '1.0.0-alpha.1', '1.0.0-alpha.beta',
        '1.0.0-beta', '1.0.0-9', '1.0.0-0a',
        '24.4.3001-20191109021931-daa7c04131f5'
    ]

    self.assertListEqual([
        '00000001.00000000.00000000-1beta.00000002',
        '00000001.00000000.00000000-1beta.00000002',
        '00000001.00000000.00000000-1beta.00000011',
        '00000001.00000000.00000000-1rc.00000001',
        '00000001.00000000.00000000-zzzzzzzzzzzzzzzz',
        '00000001.00000000.00000000-1alpha',
        '00000001.00000000.00000000-1alpha.00000001',
        '00000001.00000000.00000000-1alpha.1beta',
        '00000001.00000000.00000000-1beta',
        '00000001.00000000.00000000-00000009', '00000001.00000000.00000000-10a',
        '00000024.00000004.00003001-120191109021931-daa7c04131f5'
    ], [semver_index.normalize(version) for version in versions])

  def test_sort(self):
    """Test sorting."""
    versions = [
        '1.0.2', '1.0.11', '1.9.0', '1.11.0',
        '1.0.1-20191109021931-daa7c04131f5',
        '1.0.1-pre.0.20191109021931-daa7c04131f5',
        '1.0.1-0.20191109021931-daa7c04131f5', '1.0.1-beta.2', '1.0.1-beta.11',
        '1.0.1-rc.1', '1.0.1', '1.0.1-alpha', '1.0.1-alpha.1',
        '1.0.1-alpha.beta', '1.0.1-beta', '1.0.1-9', '1.0.1-0a'
    ]

    self.assertListEqual([
        '1.0.1-0.20191109021931-daa7c04131f5',
        '1.0.1-9',
        '1.0.1-0a',
        '1.0.1-20191109021931-daa7c04131f5',
        '1.0.1-alpha',
        '1.0.1-alpha.1',
        '1.0.1-alpha.beta',
        '1.0.1-beta',
        '1.0.1-beta.2',
        '1.0.1-beta.11',
        '1.0.1-pre.0.20191109021931-daa7c04131f5',
        '1.0.1-rc.1',
        '1.0.1',
        '1.0.2',
        '1.0.11',
        '1.9.0',
        '1.11.0',
    ], sorted(versions, key=semver_index.normalize))

    # Sanity check that the python semver library agrees.
    self.assertListEqual(
        sorted(versions, key=semver.Version.parse),
        sorted(versions, key=semver_index.normalize))


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 OSV Schema Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GHSA to OSV converter."""
import argparse
import json
import os
import re
import traceback
from typing import Any, Dict, Optional, List

import osv
import osv.ecosystems

# GHSA ecosystem -> OSV ecosystem.
ECOSYSTEM_MAP = {
    'NPM': 'npm',
    'GO': 'Go',
    'MAVEN': 'Maven',
    'PIP': 'PyPI',
    'RUBYGEMS': 'RubyGems',
    'NUGET': 'NuGet',
    'COMPOSER': 'Packagist',
    'RUST': 'crates.io',
    'ACTIONS': 'GitHub Actions',
}

NAME_NORMALIZER = {
    # Per https://www.python.org/dev/peps/pep-0503/#normalized-names.
    'PyPI': lambda name: re.sub(r'[-_.]+', '-', name).lower()
}

SEMVER_ECOSYSTEMS = {
    'npm',
    'Go',
}


class GhsaVersionSpec:
    """GHSA version spec."""
    operator: str
    version: str

    def __init__(self, operator: str, version: str):
        self.operator = operator
        self.version = version


class GhsaRange:
    """GHSA Range."""
    lower: Optional[GhsaVersionSpec] = None
    upper: Optional[GhsaVersionSpec] = None
    exact: Optional[GhsaVersionSpec] = None


def parse_ghsa_range(ghsa_range: str) -> GhsaRange:
    """Parses a GHSA version range."""
    # GHSA range format is described at:
    # https://docs.github.com/en/graphql/reference/objects#securityvulnerability
    # "= 0.2.0" denotes a single vulnerable version.
    # "<= 1.0.8" denotes a version range up to and including the specified
    # version
    # "< 0.1.11" denotes a version range up to, but excluding, the specified
    # version
    # ">= 4.3.0, < 4.3.5" denotes a version range with a known minimum and
    # maximum version.
    # ">= 0.0.1" denotes a version range with a known minimum, but no known
    # maximum.
    # (Undocumented) ">" is also a valid operator.
    parts = [part.strip() for part in ghsa_range.split(',')]

    parsed_range = GhsaRange()
    for part in parts:
        try:
            operator, version = part.split()
        except ValueError as exc:
            raise ValueError('Failed to parse GHSA range: '
                             f'"{ghsa_range}"') from exc

        spec = GhsaVersionSpec(operator, version)

        if operator == '=':
            parsed_range.exact = spec
        elif operator in ('>=', '>'):
            parsed_range.lower = spec
        elif operator in ('<=', '<'):
            parsed_range.upper = spec
        else:
            raise ValueError(f'Unknown operator "{operator}"')

    if parsed_range.exact and (parsed_range.lower or parsed_range.upper):
        raise ValueError('Range with both exact and lower/upper bounds: '
                         f'"{ghsa_range}"')

    return parsed_range


def convert_file(input_path: str, output_path: str):
    """Converts `input_path` from GHSA JSON and output OSV JSON at
    `output_path`."""
    with open(input_path) as handle:
        ghsa = json.load(handle)

    entry = convert(ghsa)
    vuln = osv.parse_vulnerability_from_dict(entry)

    osv.analyze(vuln,
                analyze_git=False,
                detect_cherrypicks=False,
                versions_from_repo=False)

    osv.write_vulnerability(vuln, output_path)


def convert_reference(reference: Dict[str, str]) -> Dict[str, str]:
    """Converts a GHSA reference to an OSV reference."""
    ref_type = 'WEB'

    if 'github.com/advisories/' in reference['url']:
        ref_type = 'ADVISORY'

    if 'nvd.nist.gov/vuln/detail/' in reference['url']:
        ref_type = 'ADVISORY'

    return {
        'type': ref_type,
        'url': reference['url'],
    }


def convert(ghsa: Dict[str, Any]) -> Dict[str, Any]:
    """Converts a GHSA entry to an OSV entry."""
    entry = {
        'schema_version': '1.5.0',
        'id': ghsa['ghsaId'],
        'aliases': [
            val['value'] for val in ghsa['identifiers']
            if val['value'] != ghsa['ghsaId']
        ],
        'published': ghsa['publishedAt'],
        'modified': ghsa['updatedAt'],
    }

    # Split up the dict assignments to preserve order of date related fields.
    withdrawn = ghsa.get('withdrawnAt')
    if withdrawn:
        entry['withdrawn'] = withdrawn

    entry.update({
        'summary': ghsa['summary'],
        'details': ghsa['description'],
        'references': [convert_reference(ref) for ref in ghsa['references']]
    })

    entry['affected'] = get_affected(ghsa)
    return entry


def get_affected(ghsa: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Converts the GHSA entry into an OSV "affected" entry."""
    package_to_vulns = {}

    # Group vulnerabilities by (ecosystem, package).
    for vuln in ghsa.get('vulnerabilities', []).get('nodes', []):
        package = vuln['package']
        mapped_ecosystem = ECOSYSTEM_MAP[package['ecosystem']]
        package_to_vulns.setdefault((mapped_ecosystem, package['name']),
                                    []).append(vuln)

    cwes = ghsa.get('cwes', {}).get('nodes', [])

    # Convert the grouped vulnerabilities in OSV range structures.
    affected = []
    for (ecosystem, name), vulns in package_to_vulns.items():
        if ecosystem in NAME_NORMALIZER:
            name = NAME_NORMALIZER[ecosystem](name)

        current = {
            'package': {
                'ecosystem': ecosystem,
                'name': name,
            },
            'ranges': [],
            'versions': [],
            'database_specific': {
                # Attribution.
                'ghsa': ghsa['permalink'],
                'cwes': cwes,
            },
        }
        affected.append(current)

        current_range = {
            'type': 'SEMVER' if ecosystem in SEMVER_ECOSYSTEMS else 'ECOSYSTEM',
            'events': [],
        }
        current_events = current_range['events']

        affects_all_prior = False

        for vuln in vulns:
            first_patched = vuln.get('firstPatchedVersion')
            if first_patched:
                first_patched = first_patched.get('identifier')

            ghsa_range = parse_ghsa_range(vuln.get('vulnerableVersionRange'))

            if ghsa_range.exact:
                if first_patched:
                    # A patch is specified, so convert it to an OSV range.
                    current_events.extend([
                        {
                            'introduced': ghsa_range.exact.version
                        },
                        {
                            'fixed': first_patched
                        },
                    ])
                else:
                    # No patch, so just add it to the explicit versions array.
                    current['versions'].append(ghsa_range.exact.version)

                continue

            if ghsa_range.lower:
                if ghsa_range.lower.operator == '>=':
                    current_events.append(
                        {'introduced': ghsa_range.lower.version})
                elif ghsa_range.lower.operator == '>':
                    ecosystem_helper = osv.ecosystems.get(ecosystem)
                    if not ecosystem_helper:
                        raise ValueError(
                            f'> is not supported for {ecosystem} yet')

                    current_events.append({
                        'introduced':
                        ecosystem_helper.next_version(name,
                                                      ghsa_range.lower.version)
                    })
            else:
                affects_all_prior = True

            if ghsa_range.upper:
                if ghsa_range.upper.operator == '<=':
                    if first_patched:
                        # "fixed" events are preferred over "last_affected"
                        current_events.append({'fixed': first_patched})
                    else:
                        current_events.append({'last_affected': ghsa_range.upper.version})
                elif ghsa_range.upper.operator == '<':
                    current_events.append({'fixed': ghsa_range.upper.version})
            elif first_patched:
                # No upper bound set in the range, check the firstPatchedVersion.
                current_events.append({'fixed': first_patched})

        if affects_all_prior:
            current_events.insert(0, {'introduced': '0'})

        if current_events:
            # Only add the range if there is at least one event.
            current['ranges'].append(current_range)

    return affected


def main():
    """Main function."""
    parser = argparse.ArgumentParser(description='GHSA to OSV converter.')
    parser.add_argument('input_files', help='Input files', nargs='+')
    parser.add_argument('-o',
                        '--output-dir',
                        help='Output directory',
                        required=True)

    args = parser.parse_args()

    for input_path in args.input_files:
        try:
            convert_file(
                input_path,
                os.path.join(args.output_dir, os.path.basename(input_path)))
        except Exception:
            print('Failed to convert', input_path)
            traceback.print_exc()


if __name__ == '__main__':
    main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bug helper tests."""

import unittest

from . import purl_helpers


class PurlHelpersTest(unittest.TestCase):
  """purl_helpers tests."""

  def tests_package_to_purl(self):
    """Test PURL generation."""
    self.assertEqual('pkg:rpm/almalinux/libvpx',
                     purl_helpers.package_to_purl('AlmaLinux', 'libvpx'))

    self.assertEqual('pkg:apk/alpaquita/xz?arch=source',
                     purl_helpers.package_to_purl('Alpaquita', 'xz'))

    self.assertEqual(
        'pkg:apk/bellsoft-hardened-containers/jq?arch=source',
        purl_helpers.package_to_purl('BellSoft Hardened Containers', 'jq'))

    self.assertEqual('pkg:apk/alpine/postgresql14?arch=source',
                     purl_helpers.package_to_purl('Alpine', 'postgresql14'))

    self.assertEqual('pkg:apk/chainguard/solr',
                     purl_helpers.package_to_purl('Chainguard', 'solr'))

    self.assertEqual('pkg:cran/commonmark',
                     purl_helpers.package_to_purl('CRAN', 'commonmark'))

    self.assertEqual('pkg:conan/openssl',
                     purl_helpers.package_to_purl('ConanCenter', 'openssl'))

    self.assertEqual('pkg:pypi/django',
                     purl_helpers.package_to_purl('PyPI', 'django'))

    self.assertEqual('pkg:rpm/mageia/python-aiohttp',
                     purl_helpers.package_to_purl('Mageia', 'python-aiohttp'))

    self.assertEqual(
        'pkg:maven/org.apache.struts/struts2-core',
        purl_helpers.package_to_purl('Maven', 'org.apache.struts:struts2-core'))

    self.assertEqual('pkg:apk/minimos/nginx',
                     purl_helpers.package_to_purl('MinimOS', 'nginx'))

    self.assertEqual('pkg:npm/%40hapi/hoek',
                     purl_helpers.package_to_purl('npm', '@hapi/hoek'))

    self.assertEqual('pkg:deb/debian/nginx?arch=source',
                     purl_helpers.package_to_purl('Debian', 'nginx'))

    self.assertEqual('pkg:deb/echo/nginx',
                     purl_helpers.package_to_purl('Echo', 'nginx'))

    self.assertEqual('pkg:apk/alpine/nginx?arch=source',
                     purl_helpers.package_to_purl('Alpine', 'nginx'))

    self.assertEqual('pkg:pub/characters',
                     purl_helpers.package_to_purl('Pub', 'characters'))

    self.assertEqual(
        'pkg:swift/github.com/Alamofire/Alamofire',
        purl_helpers.package_to_purl('SwiftURL',
                                     'github.com/Alamofire/Alamofire'))

    # Reverted tests based on test_parse_purl (without parse_purl calls)
    self.assertEqual('pkg:bitnami/moodl',
                     purl_helpers.package_to_purl('Bitnami', 'moodl'))

    self.assertEqual('pkg:cargo/surrealdb',
                     purl_helpers.package_to_purl('crates.io', 'surrealdb'))

    self.assertEqual(
        'pkg:golang/github.com/treeverse/lakefs',
        purl_helpers.package_to_purl('Go', 'github.com/treeverse/lakefs'))

    self.assertEqual('pkg:hackage/process',
                     purl_helpers.package_to_purl('Hackage', 'process'))

    self.assertEqual('pkg:hex/test-package',
                     purl_helpers.package_to_purl('Hex', 'test-package'))

    self.assertEqual('pkg:hex/acme/foo',
                     purl_helpers.package_to_purl('Hex', 'acme/foo'))

    self.assertEqual('pkg:npm/test-package',
                     purl_helpers.package_to_purl('npm', 'test-package'))

    self.assertEqual('pkg:nuget/test-package',
                     purl_helpers.package_to_purl('NuGet', 'test-package'))

    self.assertEqual('pkg:rpm/openeuler/test-package',
                     purl_helpers.package_to_purl('openEuler', 'test-package'))

    self.assertEqual('pkg:rpm/opensuse/test-package',
                     purl_helpers.package_to_purl('openSUSE', 'test-package'))

    self.assertEqual('pkg:generic/test-package',
                     purl_helpers.package_to_purl('OSS-Fuzz', 'test-package'))

    self.assertEqual(
        'pkg:composer/spencer14420/sp-php-email-handler',
        purl_helpers.package_to_purl('Packagist',
                                     'spencer14420/sp-php-email-handler'))

    self.assertEqual('pkg:pub/test-package',
                     purl_helpers.package_to_purl('Pub', 'test-package'))

    self.assertEqual('pkg:rpm/redhat/test-package',
                     purl_helpers.package_to_purl('Red Hat', 'test-package'))

    self.assertEqual(
        'pkg:rpm/rocky-linux/test-package',
        purl_helpers.package_to_purl('Rocky Linux', 'test-package'))

    self.assertEqual('pkg:gem/test-package',
                     purl_helpers.package_to_purl('RubyGems', 'test-package'))

    self.assertEqual('pkg:rpm/suse/test-package',
                     purl_helpers.package_to_purl('SUSE', 'test-package'))

    self.assertEqual(
        'pkg:swift/github.com/shareup/wasm-interpreter-apple',
        purl_helpers.package_to_purl(
            'SwiftURL', 'github.com/shareup/wasm-interpreter-apple'))

    self.assertEqual('pkg:deb/ubuntu/pygments',
                     purl_helpers.package_to_purl('Ubuntu', 'pygments'))

    self.assertEqual('pkg:apk/wolfi/test-package',
                     purl_helpers.package_to_purl('Wolfi', 'test-package'))

    self.assertIsNone(purl_helpers.package_to_purl('Invalid', 'test-package'))

  def test_parse_purl(self):
    """Test parse purl"""
    self.assertEqual(('AlmaLinux', 'libvpx', None),
                     purl_helpers.parse_purl('pkg:rpm/almalinux/libvpx'))

    self.assertEqual(('Alpaquita', 'xz', None),
                     purl_helpers.parse_purl(
                         'pkg:apk/alpaquita/xz?arch=source&distro=stream'))

    self.assertEqual(
        ('Alpine', 'postgresql14', None),
        purl_helpers.parse_purl('pkg:apk/alpine/postgresql14?arch=source'))

    self.assertEqual(('Bitnami', 'moodl', None),
                     purl_helpers.parse_purl('pkg:bitnami/moodl'))

    self.assertEqual(('Chainguard', 'solr', None),
                     purl_helpers.parse_purl('pkg:apk/chainguard/solr'))

    self.assertEqual(('crates.io', 'surrealdb', '2.1.0'),
                     purl_helpers.parse_purl('pkg:cargo/surrealdb@2.1.0'))

    self.assertEqual(('CRAN', 'commonmark', None),
                     purl_helpers.parse_purl('pkg:cran/commonmark'))

    self.assertEqual(('ConanCenter', 'openssl', '3.0.3'),
                     purl_helpers.parse_purl('pkg:conan/openssl@3.0.3'))

    self.assertEqual(('Debian', 'mpg123', '1.26.4-1+deb11u1'),
                     purl_helpers.parse_purl(
                         'pkg:deb/debian/mpg123@1.26.4-1+deb11u1?arch=source'))

    self.assertEqual(('Go', 'github.com/treeverse/lakefs', '1.33.0'),
                     purl_helpers.parse_purl(
                         'pkg:golang/github.com/treeverse/lakefs@1.33.0'))

    self.assertEqual(
        ('Go', 'github.com/jackc/pgx/v5', 'v5.6.0'),
        purl_helpers.parse_purl('pkg:golang/github.com/jackc/pgx@v5.6.0#v5'))

    self.assertEqual(('Hackage', 'process', None),
                     purl_helpers.parse_purl('pkg:hackage/process'))

    self.assertEqual(('Echo', 'nginx', None),
                     purl_helpers.parse_purl('pkg:deb/echo/nginx?arch=source'))

    self.assertEqual(
        ('BellSoft Hardened Containers', 'jq', None),
        purl_helpers.parse_purl(
            'pkg:apk/bellsoft-hardened-containers/jq?arch=source&distro=23'))

    self.assertEqual(('Hex', 'test-package', None),
                     purl_helpers.parse_purl('pkg:hex/test-package'))

    self.assertEqual(('Hex', 'acme/foo', '2.3.'),
                     purl_helpers.parse_purl('pkg:hex/acme/foo@2.3.'))

    self.assertEqual(('Mageia', 'python-aiohttp', None),
                     purl_helpers.parse_purl(
                         'pkg:rpm/mageia/python-aiohttp?distro=mageia-9'))

    self.assertEqual(('Maven', 'org.apache.struts:struts2-core', '1.0.0'),
                     purl_helpers.parse_purl(
                         'pkg:maven/org.apache.struts/struts2-core@1.0.0'))

    self.assertEqual(('MinimOS', 'nginx', None),
                     purl_helpers.parse_purl('pkg:apk/minimos/nginx'))

    self.assertEqual(('npm', '@hapi/hoek', '1.2.3'),
                     purl_helpers.parse_purl('pkg:npm/%40hapi/hoek@1.2.3'))

    self.assertEqual(('npm', 'test-package', None),
                     purl_helpers.parse_purl('pkg:npm/test-package'))

    self.assertEqual(('NuGet', 'test-package', '1.2.3'),
                     purl_helpers.parse_purl('pkg:nuget/test-package@1.2.3'))

    self.assertEqual(
        ('openEuler', 'test-package', '1.2.3'),
        purl_helpers.parse_purl('pkg:rpm/openeuler/test-package@1.2.3'))

    self.assertEqual(
        ('openSUSE', 'test-package', '1.2.3'),
        purl_helpers.parse_purl('pkg:rpm/opensuse/test-package@1.2.3'))

    self.assertEqual(('OSS-Fuzz', 'test-package', None),
                     purl_helpers.parse_purl('pkg:generic/test-package'))

    self.assertEqual(
        ('Packagist', 'spencer14420/sp-php-email-handler', '1.2.3'),
        purl_helpers.parse_purl(
            'pkg:composer/spencer14420/sp-php-email-handler@1.2.3'))

    self.assertEqual(('Pub', 'test-package', '1.2.3'),
                     purl_helpers.parse_purl('pkg:pub/test-package@1.2.3'))

    self.assertEqual(('PyPI', 'test-package', '1.2.3'),
                     purl_helpers.parse_purl('pkg:pypi/test-package@1.2.3'))

    self.assertEqual(
        ('Red Hat', 'test-package', '1.2.3'),
        purl_helpers.parse_purl('pkg:rpm/redhat/test-package@1.2.3'))

    self.assertEqual(
        ('Rocky Linux', 'test-package', '1.2.3'),
        purl_helpers.parse_purl('pkg:rpm/rocky-linux/test-package@1.2.3'))

    self.assertEqual(('RubyGems', 'test-package', '1.2.3'),
                     purl_helpers.parse_purl('pkg:gem/test-package@1.2.3'))

    self.assertEqual(('SUSE', 'test-package', '1.2.3'),
                     purl_helpers.parse_purl('pkg:rpm/suse/test-package@1.2.3'))

    self.assertEqual(
        ('SwiftURL', 'github.com/shareup/wasm-interpreter-apple', None),
        purl_helpers.parse_purl(
            'pkg:swift/github.com/shareup/wasm-interpreter-apple'))

    self.assertEqual(('Ubuntu', 'pygments', '2.11.2+dfsg-2ubuntu0.1'),
                     purl_helpers.parse_purl(
                         'pkg:deb/ubuntu/pygments@2.11.2+dfsg-2ubuntu0.1'))

    self.assertEqual(
        ('Wolfi', 'test-package', '1.2.3'),
        purl_helpers.parse_purl('pkg:apk/wolfi/test-package@1.2.3'))

    self.assertIsNone(purl_helpers.parse_purl('pkg:bad/ubuntu/pygments'))

    with self.assertRaises(ValueError):
      purl_helpers.parse_purl('purl:apk/wolfi/test-package@1.2.3')

    self.assertEqual(
        'Alpine',
        purl_helpers.parse_purl('pkg:apk/alpine/postgresql14').ecosystem)

    self.assertEqual('moodl',
                     purl_helpers.parse_purl('pkg:bitnami/moodl').package)

    self.assertEqual(
        '2.11.2',
        purl_helpers.parse_purl('pkg:deb/ubuntu/pygments@2.11.2').version)


if __name__ == '__main__':
  unittest.main()



"""Tests for converting a CSAF document to OSV format"""
import unittest
from datetime import datetime
import json
from redhat_osv.osv import DATE_FORMAT, RedHatConverter


class TestRedHatConverter(unittest.TestCase):
    """Test end-to-end convertion from RedHAt CSAF to OSV format"""

    test_advisories = ["2024_4546", "2024_6220"]

    def test_convert_redhat(self):
        for test_advisory in self.test_advisories:
            """ Test a single demo CSAF file """
            modified_time = datetime.strptime("2024-09-02T14:30:00",
                                              "%Y-%m-%dT%H:%M:%S")
            csaf_file = f"testdata/CSAF/rhsa-{test_advisory}.json"
            expected_file = f"testdata/OSV/RHSA-{test_advisory}.json"

            with open(csaf_file, "r", encoding="utf-8") as fp:
                csaf_data = fp.read()
            converter = RedHatConverter()
            osv_data = converter.convert(csaf_data,
                                         modified_time.strftime(DATE_FORMAT))

            advisory_id = test_advisory.replace("_", ":")
            assert osv_data[0] == f"RHSA-{advisory_id}"
            result_data = json.loads(osv_data[1])

            with open(expected_file, "r", encoding="utf-8") as fp:
                expected_data = json.load(fp)
            assert expected_data == result_data


if __name__ == '__main__':
    unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Packagist ecosystem helper tests."""

import os
import vcr.unittest

from .. import ecosystems


class PackagistEcosystemTest(vcr.unittest.VCRTestCase):
  """Packagist ecosystem helper tests."""
  _TEST_DATA_DIR = os.path.join(
      os.path.dirname(os.path.abspath(__file__)), 'testdata')

  def test_packagist(self):
    """Test Packagist."""
    ecosystem = ecosystems.get('Packagist')
    # Any invalid versions will be handled.
    self.assertLess(ecosystem.sort_key('invalid'), ecosystem.sort_key('0'))
    self.assertLess(
        ecosystem.sort_key('4.3-2RC1'), ecosystem.sort_key('4.3-2RC2'))
    self.assertGreater(
        ecosystem.sort_key('4.3-2RC2'), ecosystem.sort_key('4.3-2beta5'))
    self.assertGreater(
        ecosystem.sort_key('4.3-2'), ecosystem.sort_key('4.3-2beta1'))
    self.assertGreater(ecosystem.sort_key('1.0.0'), ecosystem.sort_key('1.0'))
    self.assertEqual(
        ecosystem.sort_key('1.0.0rc2'), ecosystem.sort_key('1.0.0.rc2'))

    # Check >= / <= methods
    self.assertGreaterEqual(
        ecosystem.sort_key('1.10-2RC1'), ecosystem.sort_key('1.2-2RC1'))
    self.assertLessEqual(
        ecosystem.sort_key('1.2-2RC1'), ecosystem.sort_key('1.10-2RC1'))

    enumerated_versions = ecosystem.enumerate_versions('neos/neos', '3.3.0',
                                                       '4.4.0')
    self.assertIn('4.3.19', enumerated_versions)
    self.assertIn('4.2.18', enumerated_versions)
    self.assertIn('3.3.1', enumerated_versions)
    self.assertIn('3.3.0', enumerated_versions)

    with open(os.path.join(self._TEST_DATA_DIR,
                           'packagist_test_cases.txt')) as file:
      for line in file.readlines():
        if line.startswith('//') or line.isspace():
          continue
        pieces = line.strip('\n').split(' ')
        sort_value = ecosystem.sort_key(pieces[0]).__cmp__(
            ecosystem.sort_key(pieces[2]))

        if pieces[1] == '<':
          expected_value = -1
        elif pieces[1] == '=':
          expected_value = 0
        elif pieces[1] == '>':
          expected_value = 1
        else:
          raise RuntimeError('Input not expected: ' + pieces[1])

        self.assertEqual(expected_value, sort_value, pieces)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Caching interface and implementations"""
import datetime
import functools
import inspect
import json
import typing


class Cache:
  """Cache Interface"""

  def get(self, key):
    raise NotImplementedError

  def set(self, key, value, ttl):
    raise NotImplementedError


class _CacheEntry:
  data: typing.Any
  # TODO(rexpan):
  #  Add more complex expiry logic by checking Last-Modified headers
  expiry: float

  def __init__(self, data, ttl):
    self.data = data
    self.expiry = datetime.datetime.now().timestamp() + ttl


def _check_json_serializable(obj):
  """Raise exception if `obj` is not JSON serializable."""
  # Cache implementations require keys and values to be JSON serializable.
  json.dumps(obj)


class InMemoryCache(Cache):
  """In memory cache implementation."""

  key_val_map: typing.Dict[str, _CacheEntry]

  def __init__(self):
    self.key_val_map = {}

  def get(self, key):
    entry = self.key_val_map.get(key)
    if not entry:
      return None

    if entry.expiry >= datetime.datetime.now().timestamp():
      return entry.data

    self.key_val_map.pop(key)
    return None

  def set(self, key, value, ttl):
    self.key_val_map[key] = _CacheEntry(value, ttl)


def cached(cache: Cache, ttl: int = 60 * 60):
  """Function decorator to cache results.

  Args:
    cache: Cache object to store information, each function also has a unique
      key for values cached, therefore you can safely share the same cache
      object across different functions.
    ttl: Time to live in seconds (default 1 hour).
  """

  def decorator(func):
    # Get the name of the function decorated, this will be used as the key in
    # the cache.
    unique_f_key = ('FUNC_MODULE_NAME', inspect.getmodule(func).__name__,
                    func.__qualname__)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
      # TODO(ochang): Detect and handle `self` arguments.
      sig = inspect.signature(func)
      # Passing through the arguments made to the function
      bound_args = sig.bind(*args, **kwargs)
      # Apply the default since the cache could be external (Redis)
      # If default changes, we want a new key to be generated
      bound_args.apply_defaults()
      # Making it hashable and combining it with the function name
      cache_key = (*bound_args.arguments.items(), unique_f_key)
      _check_json_serializable(cache_key)
      cached_value = cache.get(cache_key)
      if cached_value:
        return cached_value

      # Cache miss, cache the return value from the decorated function
      value = func(*args, **kwargs)
      _check_json_serializable(value)
      cache.set(cache_key, value, ttl)

      return value

    return wrapper

  return decorator



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Analyze CLI tool."""

import argparse
import json
import logging
import os
import re
import subprocess

import yaml

# pylint: disable=relative-beyond-top-level
from .. import impact
from .. import sources
from .. import vulnerability_pb2


def main():
  logging.basicConfig(level=logging.INFO)
  parser = argparse.ArgumentParser(description='Analyze')
  parser.add_argument(
      '--analyze_git',
      help='Analyze git ranges',
      choices=['true', 'false'],
      default='true')
  parser.add_argument(
      '--detect_cherrypicks',
      help='Detect git cherry-picks (slow), implies consider_all_branches',
      choices=['true', 'false'],
      default='false')
  parser.add_argument(
      '--consider_all_branches',
      help='Consider all git branches',
      choices=['true', 'false'],
      default='false')
  parser.add_argument(
      '--format',
      help='Analyze git ranges',
      choices=['yaml', 'json'],
      default='yaml')
  parser.add_argument(
      '--key_path', help='Key path where vulnerability is stored', default='')
  parser.add_argument(
      '--checkout_path', help='Path for checking out repositories')
  parser.add_argument(
      '--pr_base', help='Pull request base branch (to diff against).')
  parser.add_argument(
      '--skip_pattern', help='Regex pattern to match skipped files.')
  parser.add_argument('paths', nargs='*', help='File paths')

  args = parser.parse_args()

  analyze_git = args.analyze_git == 'true'
  detect_cherrypicks = args.detect_cherrypicks == 'true'
  consider_all_branches = args.consider_all_branches == 'true'

  if args.pr_base:
    paths = subprocess.check_output(
        ['git', 'diff', '--name-only', args.pr_base]).decode().splitlines()
  else:
    paths = args.paths

  if not paths:
    print('No vulnerability paths specified.')

  for path in paths:
    if args.skip_pattern and re.search(args.skip_pattern, path):
      continue

    ext = os.path.splitext(path)[1]
    if args.format == 'yaml':
      if ext not in sources.YAML_EXTENSIONS:
        continue
    else:
      assert args.format == 'json'
      if ext not in sources.JSON_EXTENSIONS:
        continue

    if not os.path.exists(path):
      continue

    analyze(path, args.checkout_path, args.key_path, analyze_git,
            detect_cherrypicks, consider_all_branches)


def analyze(path, checkout_path, key_path, analyze_git, detect_cherrypicks,
            consider_all_branches):
  """Analyze and write changes to file."""
  logging.info('Analyzing %s', path)
  try:
    vuln = sources.parse_vulnerability(path, key_path)
  except Exception:
    logging.error('Failed to parse %s', path)
    return

  result = impact.analyze(
      vuln,
      analyze_git=analyze_git,
      checkout_path=checkout_path,
      detect_cherrypicks=detect_cherrypicks,
      consider_all_branches=consider_all_branches)

  if not result.has_changes:
    logging.info('No changes required.')
    return

  logging.info('Found %s affected commits.', len(result.commits))
  logging.info('Writing changes.')
  sources.write_vulnerability(vuln, path, key_path)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ubuntu ecosystem helper tests."""

import unittest
from .. import ecosystems


class UbuntuEcosystemTest(unittest.TestCase):
  """Ubuntu ecosystem helper tests."""

  def test_ubuntu(self):
    """Test sort_key"""
    ecosystem = ecosystems.get('Ubuntu')
    self.assertGreater(
        ecosystem.sort_key('2.42.8+dfsg-1ubuntu0.3'),
        ecosystem.sort_key('2.40.0+dfsg-3ubuntu0.5'))
    self.assertGreater(
        ecosystem.sort_key('2.42.8+dfsg-1ubuntu0.3'),
        ecosystem.sort_key('2.42.8+dfsg-1ubuntu0.2'))
    self.assertGreater(ecosystem.sort_key('5.4.13-1'), ecosystem.sort_key('0'))
    self.assertGreater(
        ecosystem.sort_key('5.4.13-1'), ecosystem.sort_key('3.2.30-1'))
    self.assertGreater(
        ecosystem.sort_key('invalid'), ecosystem.sort_key('3.2.30-1'))
    # Check >= / <= methods
    self.assertGreaterEqual(
        ecosystem.sort_key('2.10.8+dfsg-1ubuntu0.3'),
        ecosystem.sort_key('2.2.8+dfsg-1ubuntu0.3'))
    self.assertLessEqual(
        ecosystem.sort_key('2.2.8+dfsg-1ubuntu0.3'),
        ecosystem.sort_key('2.10.8+dfsg-1ubuntu0.3'))



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Maven ecosystem helper."""

import collections
import functools
import re

from .ecosystems_base import EnumerableEcosystem, DepsDevMixin


# pylint: disable=line-too-long
# Maven's very complicated spec:
# http://maven.apache.org/pom.html#Dependency_Version_Requirement_Specification

_TO_TRIM = ('0', '', 'final', 'ga')
_KEYWORD_ORDER = ('alpha', 'beta', 'milestone', 'rc', 'snapshot', '', 'sp')


def qualifier_order(token):
  """Returns an integer representing a token's order."""
  # ".qualifier" < "-qualifier" < "-number" < ".number"
  if token.value.isdigit():
    if token.prefix == '-':
      return 2

    assert token.prefix == '.'
    return 3

  if token.prefix == '-':
    return 1

  assert token.prefix == '.'
  return 0


class VersionToken(
    collections.namedtuple(
        'VersionToken', 'prefix value is_null', defaults=(False,))):
  """Version token."""

  __slots__ = ()

  def __eq__(self, other):
    return self.prefix == other.prefix and self.value == other.value

  def __lt__(self, other):
    if self.prefix == other.prefix:
      # if the prefix is the same, then compare the token:
      if self.value.isdigit() and other.value.isdigit():
        # Numeric tokens have the natural order.
        return int(self.value) < int(other.value)
      # The spec is unclear, but according to Maven's implementation, numerics
      # sort after non-numerics, **unless it's a null value**.
      # https://github.com/apache/maven/blob/965aaa53da5c2d814e94a41d37142d0d6830375d/maven-artifact/src/main/java/org/apache/maven/artifact/versioning/ComparableVersion.java#L443
      if self.value.isdigit() and not self.is_null:
        return False

      if other.value.isdigit() and not other.is_null:
        return True

      # Non-numeric tokens ("qualifiers") have the alphabetical order, except
      # for the following tokens which come first in _KEYWORD_ORDER.
      #
      # The spec is unclear, but according to Maven's implementation, unknown
      # qualifiers sort after known qualifiers:
      # https://github.com/apache/maven/blob/965aaa53da5c2d814e94a41d37142d0d6830375d/maven-artifact/src/main/java/org/apache/maven/artifact/versioning/ComparableVersion.java#L423
      try:
        left_idx = _KEYWORD_ORDER.index(self.value)
      except ValueError:
        left_idx = len(_KEYWORD_ORDER)

      try:
        right_idx = _KEYWORD_ORDER.index(other.value)
      except ValueError:
        right_idx = len(_KEYWORD_ORDER)

      if left_idx == len(_KEYWORD_ORDER) and right_idx == len(_KEYWORD_ORDER):
        # Both are unknown qualifiers. Just do a lexical comparison.
        return self.value < other.value

      return left_idx < right_idx

    # else ".qualifier" < "-qualifier" < "-number" < ".number"
    return qualifier_order(self) < qualifier_order(other)


@functools.total_ordering
class Version:
  """Maven version."""

  def __init__(self):
    self.tokens = []

  def __str__(self):
    result = ''
    for token in self.tokens:
      result += token.prefix + token.value

    return result

  def __eq__(self, other):
    return self.tokens == other.tokens

  def __lt__(self, other):
    for i in range(max(len(self.tokens), len(other.tokens))):
      # the shorter one padded with enough "null" values with matching prefix to
      # have the same length as the longer one. Padded "null" values depend on
      # the prefix of the other version: 0 for '.', "" for '-'
      if i >= len(self.tokens):
        if other.tokens[i].prefix == '.':
          left = VersionToken('.', '0', is_null=True)
        else:
          assert other.tokens[i].prefix == '-'
          left = VersionToken('-', '', is_null=True)
      else:
        left = self.tokens[i]

      if i >= len(other.tokens):
        if self.tokens[i].prefix == '.':
          right = VersionToken('.', '0', is_null=True)
        else:
          assert self.tokens[i].prefix == '-'
          right = VersionToken('-', '', is_null=True)
      else:
        right = other.tokens[i]

      if left == right:
        continue

      return left < right

  @classmethod
  def from_string(cls, str_version):
    """Parse a version."""
    version = Version()

    # The Maven coordinate is split in tokens between dots ('.'), hyphens ('-')
    # and transitions between digits and characters. The prefix is recorded
    # and will have effect on the order.

    # Split and keep the delimiter.
    tokens = re.split(r'([-.])', str_version)
    for i in range(0, len(tokens), 2):
      if i == 0:
        # First token has no preceding prefix.
        prefix = ''
      else:
        # Preceding prefix.
        prefix = tokens[i - 1]

      # A transition between digits and characters is equivalent to a hyphen.
      # According to Maven's implementation: any non-digit is a "character":
      # https://github.com/apache/maven/blob/965aaa53da5c2d814e94a41d37142d0d6830375d/maven-artifact/src/main/java/org/apache/maven/artifact/versioning/ComparableVersion.java#L627

      # Find instances of <digit><non-digit> or <non-digit><digit>.
      # ?= makes the regex non-consuming (needed to catch overlapping
      # transitions such as <digit><non-digit><digit>).
      # This gives an array of indices where each index is where a hyphen should be.
      transitions = [
          m.span()[0] + 1
          for m in re.finditer(r'(?=(\d[^\d]|[^\d]\d))', tokens[i])
      ]
      # Add the last index so that our algorithm to split up the current token works.
      transitions.append(len(tokens[i]))

      prev_index = 0
      for j, transition in enumerate(transitions):
        if j > 0:
          prefix = '-'

        # The spec doesn't say this, but all qualifiers are case insensitive.
        current = tokens[i][prev_index:transition].lower()
        if not current:
          # Empty tokens are replaced with "0".
          current = '0'

        # Normalize "cr" to "rc" for easier comparison since they are equal in
        # precedence.
        if current == 'cr':
          current = 'rc'

        # Also do this for 'ga', 'final' which are equivalent to empty string.
        # "release" is not part of the spec but is implemented by Maven.
        if current in ('ga', 'final', 'release'):
          current = ''

        # the "alpha", "beta" and "milestone" qualifiers can respectively be
        # shortened to "a", "b" and "m" when directly followed by a number.
        if transition != len(tokens[i]):
          if current == 'a':
            current = 'alpha'

          if current == 'b':
            current = 'beta'

          if current == 'm':
            current = 'milestone'

        if current.isdigit():
          # Remove any leading zeros.
          current = str(int(current))

        version.tokens.append(VersionToken(prefix, current))
        prev_index = transition

    # Then, starting from the end of the version, the trailing "null" values
    # (0, "", "final", "ga") are trimmed.
    i = len(version.tokens) - 1
    while i > 0:  # We always want at least one token for comparison
      if version.tokens[i].value in _TO_TRIM:
        version.tokens.pop(i)
        i -= 1
        continue

      # This process is repeated at each remaining hyphen from end to start.
      while i >= 0 and version.tokens[i].prefix != '-':
        i -= 1

      i -= 1

    return version


class Maven(DepsDevMixin, EnumerableEcosystem):
  """Maven ecosystem."""

  @property
  def deps_dev_system(self) -> str:
    return 'maven'

  def sort_key(self, version):
    """Sort key."""
    return Version.from_string(version)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions."""
    return self._deps_dev_enumerate(
        package, introduced, fixed, last_affected, limits=limits)

  @property
  def supports_comparing(self):
    return True



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import google.protobuf.timestamp_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Commit(google.protobuf.message.Message):
    """Commit reference.

    In some rare cases, this may refer to a small range of commits rather than an
    exact commit (to accomodate for automated systems). In such cases, the
    semantics are as follows:

    - If this is referring to a commit which introduces a vulnerability, then
    *any* commits in the range is assumed to cause the vulnerability.

    - If this is referring to a commit which fixes a vulnerability, then *all*
    commits in the range is assumed to cause the vulnerability.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _RepoType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _RepoTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Commit._RepoType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: Commit._RepoType.ValueType  # 0
        GIT: Commit._RepoType.ValueType  # 1

    class RepoType(_RepoType, metaclass=_RepoTypeEnumTypeWrapper):
        """The repository type."""

    UNSPECIFIED: Commit.RepoType.ValueType  # 0
    GIT: Commit.RepoType.ValueType  # 1

    REPO_TYPE_FIELD_NUMBER: builtins.int
    REPO_URL_FIELD_NUMBER: builtins.int
    COMMIT_FIELD_NUMBER: builtins.int
    repo_type: global___Commit.RepoType.ValueType
    """Required. The type of the repo."""
    repo_url: builtins.str
    """Required. The publicly accessible URL of the repo that can be directly
    passed to clone commands.
    """
    commit: builtins.str
    """Required. The commit identifier (e.g. git SHA). In some cases where the
    exact commit fails to be determined by automation, this may be a small
    range in the form "A:B" which means the commit range from A (exclusive) to
    B (inclusive).
    """
    def __init__(
        self,
        *,
        repo_type: global___Commit.RepoType.ValueType = ...,
        repo_url: builtins.str = ...,
        commit: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["commit", b"commit", "repo_type", b"repo_type", "repo_url", b"repo_url"]) -> None: ...

global___Commit = Commit

@typing.final
class Package(google.protobuf.message.Message):
    """Package information and version."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    ECOSYSTEM_FIELD_NUMBER: builtins.int
    PURL_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Required. Name of the package. Should match the name used in the package
    ecosystem (e.g. the npm package name). For C/C++ projects integrated in
    OSS-Fuzz, this is the name used for the integration.
    """
    ecosystem: builtins.str
    """Required. The ecosystem for this package. 
    For the complete list of valid ecosystem names, see
    <https://ossf.github.io/osv-schema/#affectedpackage-field>.
    """
    purl: builtins.str
    """Optional. The package URL for this package."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        ecosystem: builtins.str = ...,
        purl: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["ecosystem", b"ecosystem", "name", b"name", "purl", b"purl"]) -> None: ...

global___Package = Package

@typing.final
class Event(google.protobuf.message.Message):
    """Version events."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INTRODUCED_FIELD_NUMBER: builtins.int
    FIXED_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    LAST_AFFECTED_FIELD_NUMBER: builtins.int
    introduced: builtins.str
    """The earliest version/commit where this vulnerability
    was introduced in.
    """
    fixed: builtins.str
    """The version/commit that this vulnerability was fixed in."""
    limit: builtins.str
    """The limit to apply to the range."""
    last_affected: builtins.str
    """The last affected version."""
    def __init__(
        self,
        *,
        introduced: builtins.str = ...,
        fixed: builtins.str = ...,
        limit: builtins.str = ...,
        last_affected: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["fixed", b"fixed", "introduced", b"introduced", "last_affected", b"last_affected", "limit", b"limit"]) -> None: ...

global___Event = Event

@typing.final
class Range(google.protobuf.message.Message):
    """Affected ranges."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Range._Type.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: Range._Type.ValueType  # 0
        GIT: Range._Type.ValueType  # 1
        SEMVER: Range._Type.ValueType  # 2
        ECOSYSTEM: Range._Type.ValueType  # 3

    class Type(_Type, metaclass=_TypeEnumTypeWrapper):
        """Type of the version information."""

    UNSPECIFIED: Range.Type.ValueType  # 0
    GIT: Range.Type.ValueType  # 1
    SEMVER: Range.Type.ValueType  # 2
    ECOSYSTEM: Range.Type.ValueType  # 3

    TYPE_FIELD_NUMBER: builtins.int
    REPO_FIELD_NUMBER: builtins.int
    EVENTS_FIELD_NUMBER: builtins.int
    type: global___Range.Type.ValueType
    """Required. The type of version information."""
    repo: builtins.str
    """Required if type is GIT. The publicly accessible URL of the repo that can
    be directly passed to clone commands.
    """
    @property
    def events(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Event]:
        """Required. Version event information."""

    def __init__(
        self,
        *,
        type: global___Range.Type.ValueType = ...,
        repo: builtins.str = ...,
        events: collections.abc.Iterable[global___Event] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["events", b"events", "repo", b"repo", "type", b"type"]) -> None: ...

global___Range = Range

@typing.final
class Affected(google.protobuf.message.Message):
    """Affected versions and commits."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PACKAGE_FIELD_NUMBER: builtins.int
    RANGES_FIELD_NUMBER: builtins.int
    VERSIONS_FIELD_NUMBER: builtins.int
    ECOSYSTEM_SPECIFIC_FIELD_NUMBER: builtins.int
    DATABASE_SPECIFIC_FIELD_NUMBER: builtins.int
    SEVERITY_FIELD_NUMBER: builtins.int
    @property
    def package(self) -> global___Package:
        """Optional. Package information."""

    @property
    def ranges(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Range]:
        """Required. Range information."""

    @property
    def versions(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Optional. List of affected versions."""

    @property
    def ecosystem_specific(self) -> google.protobuf.struct_pb2.Struct:
        """Optional. JSON object holding additional information about the
        vulnerability as defined by the ecosystem for which the record applies.
        """

    @property
    def database_specific(self) -> google.protobuf.struct_pb2.Struct:
        """Optional. JSON object holding additional information about the
        vulnerability as defined by the database for which the record applies.
        """

    @property
    def severity(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Severity]:
        """Optional. Severity of the vulnerability for this package."""

    def __init__(
        self,
        *,
        package: global___Package | None = ...,
        ranges: collections.abc.Iterable[global___Range] | None = ...,
        versions: collections.abc.Iterable[builtins.str] | None = ...,
        ecosystem_specific: google.protobuf.struct_pb2.Struct | None = ...,
        database_specific: google.protobuf.struct_pb2.Struct | None = ...,
        severity: collections.abc.Iterable[global___Severity] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["database_specific", b"database_specific", "ecosystem_specific", b"ecosystem_specific", "package", b"package"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["database_specific", b"database_specific", "ecosystem_specific", b"ecosystem_specific", "package", b"package", "ranges", b"ranges", "severity", b"severity", "versions", b"versions"]) -> None: ...

global___Affected = Affected

@typing.final
class Severity(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Severity._Type.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: Severity._Type.ValueType  # 0
        CVSS_V3: Severity._Type.ValueType  # 1
        CVSS_V2: Severity._Type.ValueType  # 2
        CVSS_V4: Severity._Type.ValueType  # 3
        Ubuntu: Severity._Type.ValueType  # 4

    class Type(_Type, metaclass=_TypeEnumTypeWrapper):
        """Type of the severity."""

    UNSPECIFIED: Severity.Type.ValueType  # 0
    CVSS_V3: Severity.Type.ValueType  # 1
    CVSS_V2: Severity.Type.ValueType  # 2
    CVSS_V4: Severity.Type.ValueType  # 3
    Ubuntu: Severity.Type.ValueType  # 4

    TYPE_FIELD_NUMBER: builtins.int
    SCORE_FIELD_NUMBER: builtins.int
    type: global___Severity.Type.ValueType
    """The type of this severity entry."""
    score: builtins.str
    """The quantitative score."""
    def __init__(
        self,
        *,
        type: global___Severity.Type.ValueType = ...,
        score: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["score", b"score", "type", b"type"]) -> None: ...

global___Severity = Severity

@typing.final
class Credit(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Credit._Type.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: Credit._Type.ValueType  # 0
        OTHER: Credit._Type.ValueType  # 1
        FINDER: Credit._Type.ValueType  # 2
        REPORTER: Credit._Type.ValueType  # 3
        ANALYST: Credit._Type.ValueType  # 4
        COORDINATOR: Credit._Type.ValueType  # 5
        REMEDIATION_DEVELOPER: Credit._Type.ValueType  # 6
        REMEDIATION_REVIEWER: Credit._Type.ValueType  # 7
        REMEDIATION_VERIFIER: Credit._Type.ValueType  # 8
        TOOL: Credit._Type.ValueType  # 9
        SPONSOR: Credit._Type.ValueType  # 10

    class Type(_Type, metaclass=_TypeEnumTypeWrapper): ...
    UNSPECIFIED: Credit.Type.ValueType  # 0
    OTHER: Credit.Type.ValueType  # 1
    FINDER: Credit.Type.ValueType  # 2
    REPORTER: Credit.Type.ValueType  # 3
    ANALYST: Credit.Type.ValueType  # 4
    COORDINATOR: Credit.Type.ValueType  # 5
    REMEDIATION_DEVELOPER: Credit.Type.ValueType  # 6
    REMEDIATION_REVIEWER: Credit.Type.ValueType  # 7
    REMEDIATION_VERIFIER: Credit.Type.ValueType  # 8
    TOOL: Credit.Type.ValueType  # 9
    SPONSOR: Credit.Type.ValueType  # 10

    NAME_FIELD_NUMBER: builtins.int
    CONTACT_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name to give credit to."""
    type: global___Credit.Type.ValueType
    """Optional. The type or role of the individual or entity being credited."""
    @property
    def contact(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Contact methods (URLs)."""

    def __init__(
        self,
        *,
        name: builtins.str = ...,
        contact: collections.abc.Iterable[builtins.str] | None = ...,
        type: global___Credit.Type.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["contact", b"contact", "name", b"name", "type", b"type"]) -> None: ...

global___Credit = Credit

@typing.final
class Reference(google.protobuf.message.Message):
    """Reference URL."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Reference._Type.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NONE: Reference._Type.ValueType  # 0
        """Next ID = 12."""
        ADVISORY: Reference._Type.ValueType  # 2
        ARTICLE: Reference._Type.ValueType  # 6
        DETECTION: Reference._Type.ValueType  # 9
        DISCUSSION: Reference._Type.ValueType  # 8
        EVIDENCE: Reference._Type.ValueType  # 7
        FIX: Reference._Type.ValueType  # 4
        GIT: Reference._Type.ValueType  # 11
        INTRODUCED: Reference._Type.ValueType  # 10
        PACKAGE: Reference._Type.ValueType  # 5
        REPORT: Reference._Type.ValueType  # 3
        WEB: Reference._Type.ValueType  # 1

    class Type(_Type, metaclass=_TypeEnumTypeWrapper): ...
    NONE: Reference.Type.ValueType  # 0
    """Next ID = 12."""
    ADVISORY: Reference.Type.ValueType  # 2
    ARTICLE: Reference.Type.ValueType  # 6
    DETECTION: Reference.Type.ValueType  # 9
    DISCUSSION: Reference.Type.ValueType  # 8
    EVIDENCE: Reference.Type.ValueType  # 7
    FIX: Reference.Type.ValueType  # 4
    GIT: Reference.Type.ValueType  # 11
    INTRODUCED: Reference.Type.ValueType  # 10
    PACKAGE: Reference.Type.ValueType  # 5
    REPORT: Reference.Type.ValueType  # 3
    WEB: Reference.Type.ValueType  # 1

    TYPE_FIELD_NUMBER: builtins.int
    URL_FIELD_NUMBER: builtins.int
    type: global___Reference.Type.ValueType
    """Required. The type of the reference."""
    url: builtins.str
    """Required. The URL."""
    def __init__(
        self,
        *,
        type: global___Reference.Type.ValueType = ...,
        url: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["type", b"type", "url", b"url"]) -> None: ...

global___Reference = Reference

@typing.final
class Vulnerability(google.protobuf.message.Message):
    """A vulnerability entry.
    The protobuf representation is *NOT* stable and only used for implementing
    the JSON based API.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SCHEMA_VERSION_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    PUBLISHED_FIELD_NUMBER: builtins.int
    MODIFIED_FIELD_NUMBER: builtins.int
    WITHDRAWN_FIELD_NUMBER: builtins.int
    ALIASES_FIELD_NUMBER: builtins.int
    RELATED_FIELD_NUMBER: builtins.int
    UPSTREAM_FIELD_NUMBER: builtins.int
    SUMMARY_FIELD_NUMBER: builtins.int
    DETAILS_FIELD_NUMBER: builtins.int
    AFFECTED_FIELD_NUMBER: builtins.int
    REFERENCES_FIELD_NUMBER: builtins.int
    DATABASE_SPECIFIC_FIELD_NUMBER: builtins.int
    SEVERITY_FIELD_NUMBER: builtins.int
    CREDITS_FIELD_NUMBER: builtins.int
    schema_version: builtins.str
    """The OSV schema version."""
    id: builtins.str
    """The `id` field is a unique identifier for the vulnerability entry. It is a
    string of the format `<DB>-<ENTRYID>`, where `DB` names the database and
    `ENTRYID` is in the format used by the database. For example:
    “OSV-2020-111”, “CVE-2021-3114”, or “GHSA-vp9c-fpxx-744v”.
    """
    summary: builtins.str
    """Required. One line human readable summary for the vulnerability. It is
    recommended to keep this under 120 characters.
    """
    details: builtins.str
    """Required. Any additional human readable details for the vulnerability."""
    @property
    def published(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """The RFC3339 timestamp indicating when this entry was published."""

    @property
    def modified(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """The RFC3339 timestamp indicating when this entry was last modified."""

    @property
    def withdrawn(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Optional. The RFC3339 timestamp indicating when this entry is considered to
        be withdrawn.
        """

    @property
    def aliases(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Optional. IDs for the same vulnerability in other databases."""

    @property
    def related(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Optional. List of IDs of closely related vulnerabilities, such as the same
        problem in alternate ecosystems.
        """

    @property
    def upstream(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Optional. List of IDs of upstream vulnerabilities of a vulnerability."""

    @property
    def affected(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Affected]:
        """Required. Affected commit ranges and versions."""

    @property
    def references(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Reference]:
        """Optional. URLs to more information/advisories (including the
        scheme e.g "https://").
        """

    @property
    def database_specific(self) -> google.protobuf.struct_pb2.Struct:
        """Optional. JSON object holding additional information about the
        vulnerability as defined by the database for which the record applies.
        """

    @property
    def severity(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Severity]:
        """Optional. Severity of the vulnerability."""

    @property
    def credits(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Credit]:
        """Optional. Credits for the vulnerability."""

    def __init__(
        self,
        *,
        schema_version: builtins.str = ...,
        id: builtins.str = ...,
        published: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        modified: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        withdrawn: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        aliases: collections.abc.Iterable[builtins.str] | None = ...,
        related: collections.abc.Iterable[builtins.str] | None = ...,
        upstream: collections.abc.Iterable[builtins.str] | None = ...,
        summary: builtins.str = ...,
        details: builtins.str = ...,
        affected: collections.abc.Iterable[global___Affected] | None = ...,
        references: collections.abc.Iterable[global___Reference] | None = ...,
        database_specific: google.protobuf.struct_pb2.Struct | None = ...,
        severity: collections.abc.Iterable[global___Severity] | None = ...,
        credits: collections.abc.Iterable[global___Credit] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["database_specific", b"database_specific", "modified", b"modified", "published", b"published", "withdrawn", b"withdrawn"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["affected", b"affected", "aliases", b"aliases", "credits", b"credits", "database_specific", b"database_specific", "details", b"details", "id", b"id", "modified", b"modified", "published", b"published", "references", b"references", "related", b"related", "schema_version", b"schema_version", "severity", b"severity", "summary", b"summary", "upstream", b"upstream", "withdrawn", b"withdrawn"]) -> None: ...

global___Vulnerability = Vulnerability



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for interacting with the OSV vulnerabilities in the GCS bucket."""
import datetime
import logging
import os

from google.cloud import exceptions
from google.cloud import storage

from .vulnerability_pb2 import Vulnerability

VULN_PB_PATH = 'all/pb/'

_storage_client = None


def _get_storage_client() -> storage.Client:
  global _storage_client
  if _storage_client is None:
    _storage_client = storage.Client()
  return _storage_client


def get_osv_bucket() -> storage.Bucket:
  """Gets the osv vulnerability bucket, from OSV_VULNERABILITIES_BUCKET"""
  try:
    bucket = os.environ['OSV_VULNERABILITIES_BUCKET']
    bucket = bucket.lstrip('gs://')
  except KeyError:
    logging.error('OSV_VULNERABILITIES_BUCKET environment variable not set')
    raise
  client = _get_storage_client()
  return client.bucket(bucket)


def get_by_id(vuln_id: str) -> Vulnerability:
  """Gets the OSV record from the GCS bucket.
  Raises a google.cloud.exceptions.NotFound exception if the record is missing.
  """
  bucket = get_osv_bucket()
  pb_blob = bucket.blob(os.path.join(VULN_PB_PATH, vuln_id + '.pb'))
  try:
    return Vulnerability.FromString(pb_blob.download_as_bytes())
  except exceptions.NotFound:
    # It's not necessarily an error if not found.
    logging.info('vulnerability %s not found in GCS', vuln_id)
    raise
  except:
    logging.error('failed to download %s protobuf from GCS', vuln_id)
    raise


def get_by_id_with_generation(vuln_id: str) -> tuple[Vulnerability, int] | None:
  """Gets the OSV record, and the object's generation from the GCS bucket.
  Returns None if the record is not found.
  """
  bucket = get_osv_bucket()
  pb_blob = bucket.get_blob(os.path.join(VULN_PB_PATH, vuln_id + '.pb'))
  if pb_blob is None:
    return None
  try:
    vuln = Vulnerability.FromString(pb_blob.download_as_bytes())
    return vuln, pb_blob.generation
  except:
    logging.error('failed to download %s protobuf from GCS', vuln_id)
    raise


def upload_vulnerability(vulnerability: Vulnerability,
                         generation: int | None = None):
  """Uploads the OSV record to the GCS bucket.
  If set, checks if the existing blob's generation matches `generation`
  before uploading."""
  bucket = get_osv_bucket()
  vuln_id = vulnerability.id
  modified = vulnerability.modified.ToDatetime(datetime.UTC)
  pb_blob = bucket.blob(os.path.join(VULN_PB_PATH, vuln_id + '.pb'))
  pb_blob.custom_time = modified
  pb_blob.upload_from_string(
      vulnerability.SerializeToString(deterministic=True),
      content_type='application/octet-stream',
      if_generation_match=generation)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Many tests are ported from
# https://github.com/dart-lang/pub_semver/blob/master/test/version_test.dart
"""Pub version parser tests."""

import unittest
import vcr.unittest
import warnings

from . import pub
from .. import ecosystems


class PubVersionTest(unittest.TestCase):
  """Pub version tests."""

  def setUp(self):
    self.maxDiff = None  # pylint: disable=invalid-name

  def test_comparison(self):
    """Test version comparisons."""
    # A correctly sorted list of versions.
    versions = [
        '1.0.0-alpha', '1.0.0-alpha.1', '1.0.0-beta.2', '1.0.0-beta.11',
        '1.0.0-rc.1', '1.0.0-rc.1+build.1', '1.0.0', '1.0.0+0.3.7',
        '1.3.7+build', '1.3.7+build.2.b8f12d7', '1.3.7+build.11.e0f985a',
        '2.0.0', '2.1.0', '2.2.0', '2.11.0', '2.11.1'
    ]

    for i, a_str in enumerate(versions):
      for j, b_str in enumerate(versions):
        a = pub.Version.from_string(a_str)
        b = pub.Version.from_string(b_str)
        self.assertEqual(a < b, i < j)
        self.assertEqual(a == b, i == j)

  def test_equality(self):
    """Test version equality."""

    def check_version_equals(v1, v2):
      self.assertEqual(pub.Version.from_string(v1), pub.Version.from_string(v2))

    check_version_equals('01.2.3', '1.2.3')
    check_version_equals('1.02.3', '1.2.3')
    check_version_equals('1.2.03', '1.2.3')
    check_version_equals('1.2.3-01', '1.2.3-1')
    check_version_equals('1.2.3+01', '1.2.3+1')

  def test_ge_le(self):
    """Test version >=/<=."""
    self.assertGreaterEqual(
        pub.Version.from_string('1.10.0'), pub.Version.from_string('1.2.0'))
    self.assertLessEqual(
        pub.Version.from_string('1.2.0'), pub.Version.from_string('1.10.0'))

  def test_parse(self):
    """Test versions can be parsed."""
    pub.Version.from_string('0.0.0')
    pub.Version.from_string('12.34.56')
    pub.Version.from_string('1.2.3-alpha.1')
    pub.Version.from_string('1.2.3-x.7.z-92')
    pub.Version.from_string('1.2.3+build.1')
    pub.Version.from_string('1.2.3+x.7.z-92')
    pub.Version.from_string('1.0.0-rc-1+build-1')
    # Tests invalid versions
    pub.Version.from_string('3.4.0rc3-invalid')

  def test_empty_identifier(self):
    """Test parsing versions with empty identifiers.

    Although it's unlikely that it was intentional, SemVer 2.0.0-rc.1 does not
    explicitly disallow this case. Not sure if Pub even allows it.

    This test is probably unnecessary."""

    pub.Version.from_string('1.0.0-a..b')
    pub.Version.from_string('1.0.0-.a.b')
    pub.Version.from_string('1.0.0-a.b.')
    pub.Version.from_string('1.0.0+a..b')
    pub.Version.from_string('1.0.0+.a.b')
    pub.Version.from_string('1.0.0+a.b.')
    pub.Version.from_string('1.0.0-+')
    pub.Version.from_string('1.0.0-.+.')
    pub.Version.from_string('1.0.0-....+....')

    # Basic test for ordering.
    v_empty = pub.Version.from_string('1.0.0-a..b')
    v_number = pub.Version.from_string('1.0.0-a.0.b')
    v_str = pub.Version.from_string('1.0.0-a.a.b')
    self.assertLess(v_number, v_empty)
    self.assertLess(v_empty, v_str)

    # note(michaelkedar):
    # The implementation incorrectly assumes "1.0.0-a..b" == "1.0.0-a.-.b"
    # I have decided that this extreme edge case is not worth fixing.


class PubEcosystemTest(vcr.unittest.VCRTestCase):
  """Pub ecosystem helper tests."""

  def test_next_version(self):
    """Test next_version."""
    ecosystem = ecosystems.get('Pub')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')

      self.assertEqual('2.0.0-nullsafety.0',
                       ecosystem.next_version('pub_semver', '1.4.4'))
      self.assertEqual(
          '2.0.0', ecosystem.next_version('pub_semver', '2.0.0-nullsafety.0'))
      self.assertEqual('2.1.0', ecosystem.next_version('pub_semver', '2.0.0'))
      self.assertEqual('2.1.1', ecosystem.next_version('pub_semver', '2.1.0'))

      # Versions with pre-release and build suffixes.
      self.assertEqual('3.0.0-alpha+2',
                       ecosystem.next_version('mockito', '3.0.0-alpha'))
      self.assertEqual('3.0.0-alpha+3',
                       ecosystem.next_version('mockito', '3.0.0-alpha+2'))
      self.assertEqual('3.0.0-beta',
                       ecosystem.next_version('mockito', '3.0.0-alpha+5'))
      self.assertEqual('3.0.0', ecosystem.next_version('mockito',
                                                       '3.0.0-beta+3'))
      self.assertEqual('4.1.1+1', ecosystem.next_version('mockito', '4.1.1'))
      self.assertEqual('4.1.2', ecosystem.next_version('mockito', '4.1.1+1'))

      # Version marked as retracted (go_router 4.2.1)
      self.assertEqual('4.2.1', ecosystem.next_version('go_router', '4.2.0'))
      self.assertEqual('4.2.2', ecosystem.next_version('go_router', '4.2.1'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 OSV Schema Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Debian to OSV converter."""
import argparse
import collections
import io
import json
import os
import re
from datetime import datetime, timezone
import subprocess
import typing
from urllib import request
from enum import Enum

import dateutil.parser
import markdownify
import pandas as pd

# import osv
# import osv.ecosystems

WEBWML_SECURITY_PATH = os.path.join('english', 'security')
WEBWML_LTS_SECURITY_PATH = os.path.join('english', 'lts', 'security')
SECURITY_TRACKER_DSA_PATH = os.path.join('data', 'DSA', 'list')
SECURITY_TRACKER_DTSA_PATH = os.path.join('data', 'DTSA', 'list')
SECURITY_TRACKER_DLA_PATH = os.path.join('data', 'DLA', 'list')
DEBIAN_BASE_URL = 'https://www.debian.org'

LEADING_WHITESPACE = re.compile(r'^\s')

# e.g. [25 Apr 2022] DSA-5124-1 ffmpeg - security update
DSA_PATTERN = re.compile(r'\[(.*?)]\s*([\w-]+)\s*(.*)')

# e.g. [buster] - xz-utils 5.2.4-1+deb10u1
VERSION_PATTERN = re.compile(r'\[(.*?)]\s*-\s*([^\s]+)\s*([^\s]+)')

# TODO: Alternative is to use a xml parser here,
#  though the data is not fully compliant with the xml standard
#  It is possible to parse with an html parser however

# e.g. <define-tag moreinfo>\n Some html here \n</define-tag>
WML_DESCRIPTION_PATTERN = re.compile(
    r'<define-tag moreinfo>((?:.|\n)*)</define-tag>', re.MULTILINE)

# e.g. <define-tag report_date>2022-1-04</define-tag>
WML_REPORT_DATE_PATTERN = re.compile(
    r'<define-tag report_date>(.*)</define-tag>')

# e.g. DSA-12345-2, -2 is the extension
DSA_OR_DLA_WITH_NO_EXT = re.compile(r'd[sl]a-\d+')

NOT_AFFECTED_VERSION = '<not-affected>'

# Prefix used to identify a new date line
GIT_DATE_PREFIX = '-----'


class AdvisoryType(Enum):
  DSA = 'DSA'
  DLA = 'DLA'
  DTSA = 'DTSA'

  def __str__(self):
    return self.value


class AffectedInfo:
  """Debian version info."""
  package: str
  ranges: [str]
  fixed: str
  versions: [str]
  debian_release_version: str

  def __init__(self, version: str, package: str, fixed: str):
    self.package = package
    self.fixed = fixed
    self.debian_release_version = version

  def to_dict(self):
    return {
        'package': {
            'ecosystem': 'Debian:' + self.debian_release_version,
            'name': self.package
        },
        'ranges': [{
            'type': 'ECOSYSTEM',
            'events': [{
                'introduced': '0'
            }, {
                'fixed': self.fixed
            }]
        }],
    }

  def __repr__(self):
    return json.dumps(self, default=dumper)


class Reference:
  """OSV reference format"""

  type: str
  url: str

  def __init__(self, url_type, url):
    self.type = url_type
    self.url = url


class AdvisoryInfo:
  """Debian advisory info."""

  id: str
  summary: str
  details: str
  published: str
  modified: str
  affected: [AffectedInfo]
  aliases: [str]
  references: [Reference]

  def __init__(self, adv_id: str, summary: str, published: str):
    self.id = adv_id
    self.summary = summary
    self.affected = []
    self.aliases = []
    # Set a placeholder value for published and modified, if there is wml files
    # this will be replaced
    self.published = published
    self.modified = published
    self.details = ''
    self.references = []

  def to_dict(self):
    return self.__dict__

  def __repr__(self):
    return json.dumps(self, default=dumper)


Advisories = typing.Dict[str, AdvisoryInfo]
"""Type alias for collection of advisory info"""


def create_codename_to_version() -> typing.Dict[str, str]:
  """Returns the codename to version mapping"""
  with request.urlopen(
      'https://debian.pages.debian.net/distro-info-data/debian.csv') as csv:
    df = pd.read_csv(csv, dtype=str)
    # `series` appears to be `codename` but with all lowercase
    result = dict(zip(df['series'], df['version']))
    result['sid'] = 'unstable'
    return result


def dumper(obj):
  try:
    return obj.to_dict()
  except AttributeError:
    return obj.__dict__


def parse_security_tracker_file(advisories: Advisories,
                                security_tracker_repo: str,
                                security_tracker_path: str):
  """Parses the security tracker files into the advisories object"""

  codename_to_version = create_codename_to_version()

  with open(
      os.path.join(security_tracker_repo, security_tracker_path),
      encoding='utf-8') as file_handle:
    current_advisory = None

    # Enumerate advisories + version info from security-tracker.
    for line in file_handle:
      line = line.rstrip()
      if not line:
        continue

      if LEADING_WHITESPACE.match(line):
        # Within current advisory.
        if not current_advisory:
          raise ValueError('Unexpected tab.')

        # {CVE-XXXX-XXXX CVE-XXXX-XXXX}
        line = line.lstrip()
        if line.startswith('{'):
          advisories[current_advisory].aliases = line.strip('{}').split()
          continue

        if line.startswith('NOTE:'):
          continue

        version_match = VERSION_PATTERN.match(line)
        if not version_match:
          raise ValueError('Invalid version line: ' + line)

        release_name = version_match.group(1)
        package_name = version_match.group(2)
        fixed_ver = version_match.group(3)
        if fixed_ver != NOT_AFFECTED_VERSION:
          advisories[current_advisory].affected.append(
              AffectedInfo(codename_to_version[release_name], package_name,
                           fixed_ver))

      else:
        if line.strip().startswith('NOTE:'):
          continue

        # New advisory.
        dsa_match = DSA_PATTERN.match(line)
        if not dsa_match:
          raise ValueError('Invalid line: ' + line)

        parsed_date = dateutil.parser.parse(
            dsa_match.group(1)).isoformat() + 'Z'
        current_advisory = dsa_match.group(2)
        advisories[current_advisory] = AdvisoryInfo(current_advisory,
                                                    dsa_match.group(3),
                                                    parsed_date)


def parse_webwml_files(advisories: Advisories, webwml_repo_path: str,
                       wml_file_sub_path: str):
  """Parses the webwml file into the advisories object"""
  file_path_map = {}

  for root, _, files in os.walk(
      os.path.join(webwml_repo_path, wml_file_sub_path)):
    for file in files:
      file_path_map[file] = os.path.join(root, file)

  git_relative_paths = collections.defaultdict(list)
  # Add descriptions to advisories from wml files
  for dsa_id, advisory in advisories.items():
    # remove potential extension (e.g. DSA-12345-2, -2 is the extension)
    mapped_key_no_ext = DSA_OR_DLA_WITH_NO_EXT.findall(dsa_id.lower())[0]
    wml_path = file_path_map.get(mapped_key_no_ext + '.wml')
    data_path = file_path_map.get(mapped_key_no_ext + '.data')

    if not wml_path:
      print('No WML file yet for this: ' + mapped_key_no_ext +
            ', creating partial schema')
      continue

    with open(wml_path, encoding='iso-8859-2') as handle:
      data = handle.read()
      html = WML_DESCRIPTION_PATTERN.findall(data)[0]
      res = markdownify.markdownify(html)
      advisory.details = res

    with open(data_path, encoding='utf-8') as handle:
      data: str = handle.read()
      report_date: str = WML_REPORT_DATE_PATTERN.findall(data)[0]

      # Split by ',' here for the occasional case where there
      # are two dates in the 'publish' field.
      # Multiple dates are caused by major modification later on.
      # This is accounted for with the modified timestamp with git
      # below though, so we don't need to parse them here
      advisory.published = (
          datetime.strptime(report_date.split(',')[0], '%Y-%m-%d').isoformat() +
          'Z')

    advisory_url_path = os.path.relpath(
        wml_path, os.path.join(webwml_repo_path, 'english'))
    advisory_url_path = os.path.splitext(advisory_url_path)[0]
    advisory_url = f'{DEBIAN_BASE_URL}/{advisory_url_path}'

    advisory.references.append(Reference('ADVISORY', advisory_url))

    git_relative_path_wml = os.path.relpath(wml_path, webwml_repo_path)
    git_relative_path_data = os.path.relpath(data_path, webwml_repo_path)
    git_relative_paths[git_relative_path_wml].append(dsa_id)
    git_relative_paths[git_relative_path_data].append(dsa_id)

  modified_date_dict = collections.defaultdict(
      lambda: datetime.fromtimestamp(0, timezone.utc))
  current_date = None
  proc = subprocess.Popen([
      'git', 'log', f'--pretty={GIT_DATE_PREFIX}%aI', '--name-only',
      '--author-date-order'
  ],
                          cwd=webwml_repo_path,
                          stdout=subprocess.PIPE)
  # Loop through each commit to get the first time a file is mentioned
  # Save the date as the last modified date of said file
  for line in io.TextIOWrapper(proc.stdout, encoding='utf-8'):
    line = line.strip()
    if not line:
      continue

    if line.startswith(GIT_DATE_PREFIX):
      current_date = datetime.fromisoformat(
          line[len(GIT_DATE_PREFIX):]).astimezone(timezone.utc)
      continue

    dsa_ids = git_relative_paths.pop(line, None)
    if not dsa_ids:
      continue

    for dsa_id in dsa_ids:
      # Set modified date to the latest of the .data and .wml files.
      modified_date_dict[dsa_id] = max(modified_date_dict[dsa_id], current_date)

    # Empty dictionary means no more files need modification dates
    # Safely skip rest of the commits
    if not git_relative_paths:
      break

  for dsa_id, modified_date in modified_date_dict.items():
    # OSV spec requires a "Z" offset
    advisories[dsa_id].modified = modified_date.isoformat().replace(
        '+00:00', 'Z')


def write_output(output_dir: str, advisories: Advisories):
  """Writes the advisory dict into individual json files"""
  for dsa_id, advisory in advisories.items():
    # Skip advisories that do not affect anything
    if len(advisory.affected) == 0:
      print('Skipping: ' + dsa_id + ' because no affected versions')
      continue

    with open(
        os.path.join(output_dir, dsa_id + '.json'), 'w',
        encoding='utf-8') as output_file:
      output_file.write(json.dumps(advisory, default=dumper, indent=2))
      print(
          'Writing: ' + os.path.join(output_dir, dsa_id + '.json'), flush=True)

  print('Complete')


def is_dsa_file(name: str):
  """Check if filename is a DSA output file, e.g. DSA-1234-1.json"""
  return name.startswith('DSA-') and name.endswith('.json')


def convert_debian(webwml_repo: str, security_tracker_repo: str,
                   output_dir: str, adv_type: AdvisoryType):
  """Convert Debian advisory data into OSV."""
  advisories: Advisories = {}

  if adv_type == AdvisoryType.DLA:
    parse_security_tracker_file(advisories, security_tracker_repo,
                                SECURITY_TRACKER_DLA_PATH)
    parse_webwml_files(advisories, webwml_repo, WEBWML_LTS_SECURITY_PATH)
  elif adv_type == AdvisoryType.DSA:
    parse_security_tracker_file(advisories, security_tracker_repo,
                                SECURITY_TRACKER_DSA_PATH)
    parse_webwml_files(advisories, webwml_repo, WEBWML_SECURITY_PATH)
  elif adv_type == AdvisoryType.DTSA:
    parse_security_tracker_file(advisories, security_tracker_repo,
                                SECURITY_TRACKER_DTSA_PATH)
  else:
    raise ValueError('Invalid advisory type')

  write_output(output_dir, advisories)


def main():
  """Main function."""
  parser = argparse.ArgumentParser(description='Debian to OSV converter.')
  parser.add_argument('webwml_repo', help='Debian wml repo')
  parser.add_argument(
      'security_tracker_repo', help='Debian security-tracker repo')
  parser.add_argument(
      '-o', '--output-dir', help='Output directory', required=True)
  parser.add_argument(
      '--adv_type',
      help='Advisory type',
      type=AdvisoryType,
      choices=list(AdvisoryType))
  parser.set_defaults(feature=False)

  args = parser.parse_args()

  convert_debian(args.webwml_repo, args.security_tracker_repo, args.output_dir,
                 args.adv_type)


if __name__ == '__main__':
  main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ecosystem helper for ecosystems using SemVer."""
from warnings import deprecated

from .ecosystems_base import OrderedEcosystem
from .. import semver_index


class SemverLike(OrderedEcosystem):
  """Ecosystem helper for ecosystems that use SEMVER-compatible versioning,
  but use the ECOSYSTEM version type."""

  def sort_key(self, version):
    """Sort key."""
    try:
      return semver_index.parse(version)
    except ValueError:
      # If a user gives us an unparsable semver version,
      # treat it as a very large version so as to not match anything.
      return semver_index.parse('999999')


class SemverEcosystem(SemverLike):
  """Ecosystems which use the 'SEMVER' OSV version type"""

  @deprecated('Avoid using this method. '
              'It is provided only to maintain existing tooling.')
  def next_version(self, package, version):
    """Get the next version after the given version."""
    del package  # Unused.
    parsed_version = semver_index.parse(version)
    if parsed_version.prerelease:
      return version + '.0'

    return str(parsed_version.bump_patch()) + '-0'
